"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import numpy as np
import networkx as nx
from numba import njit, prange

from qrisp.circuit import QuantumCircuit, Qubit, PTControlledOperation, ControlledOperation, transpile, QubitAlloc, Instruction, fast_append
from qrisp.misc import get_depth_dic, retarget_instructions, reduce_depth



#The purpose of this function is to dynamically (de)allocate qubits when they are need or not needed anymore.
#The qompiler function knows when a qubit is ready to deallocate (ie. it is in |0> state) due to a gate called QubitDealloc.
#After some preprocessing steps, this function will iterate through the data and dynamically (de)allocate if it is required.
#Another feature of this function is that it replaces mcx gates where the method is set to "auto" with implementations that
#fit the situation ie. it takes into consideration how many clean/dirty ancilla qubits are available and synthesizes an
# mcx implementation based on this information.

#The parameter workspace can be given to extend the amount of qubits of the resulting QuantumCircuit by that amount.
#This extension can be helpfull because it gives the compiler more options to pick when allocating a new qubit.
#Since the allocation is based on a depth-reduction heuristic, giving more workspace results in many case in less depth circuits.

#It is however not only depth that can be improved by an increased workspace. Due to more clean/dirty ancillae beeing available,
#in many cases it is also possible to generate more efficient mcx implementations, thus also reducing the gate count.
def qompiler(qs, workspace = 0, disable_uncomputation = True, intended_measurements = [], cancel_qfts = True):
    
    if len(qs.data) == 0:
        return QuantumCircuit(0)
    
    with fast_append():
        qc = qs.copy()
        
        if not disable_uncomputation:
            local_qvs = qs.get_local_qvs()
            
            from qrisp.uncomputation.unqomp import uncompute_qc
            
            for qv in local_qvs:
                try:
                    qc = uncompute_qc(qc, qv.reg)
                except:
                    print(f"Warning: Automatic uncomputation for {qv.name} failed")
            
        if intended_measurements:
            #This function reorders the circuit such that the intended measurements can be executed as early as possible
            #additionally, any instructions that are not needed for the intended measurements are removed
            try:
                qc = measurement_reduction(qc, intended_measurements)
            except Exception as e:
                if "Unitary of operation " not in str(e):
                    raise e
    
        def reordering_transpile_predicate(op):
            if (isinstance(op, PTControlledOperation) and op.base_operation.name in ["x"]) or isinstance(op, LogicSynthGate):
                return False
            return True
    
        
        from qrisp.logic_synthesis import LogicSynthGate
        
        if cancel_qfts:
            # The first step is to cancel adjacent QFT gates, which are inverse to each other
            # This can happen alot because of the heavy use of Fourier arithmetic
            qft_transpile_predicate = lambda op : not "QFT" in op.name and reordering_transpile_predicate(op)
            
            qc = transpile(qc, transpile_predicate = qft_transpile_predicate)
            
            qc = qft_cancellation(qc)
        
        #We now reorder the transpiled QuantumCircuit. Reordering is performed based on the DAG representation
        #of Unqomp. The advantage of this representation is that it abstracts away non-trivial commutation relations
        #between permeable gates. The actual ordering is performed by performing a topological sort on this dag.
        #Contrary to unqomp, we don't use a modified version of Kahns algorithm here (for more information check 
        #the topological sort function)
    
        #The reordering process aims to find an order that minimizes that maximum amount of qubits that is needed.
        #Losely speaking it tries to reorder the data that allocations are performed as late as possible and deallocations
        #are performed as early as possible.
    
        #Note that the more aggressive we transpile, the more optimization is possible through reordering ie. less qubits.
        #By reordering we however also destroy any previous order which might have been intentionally picked to optimize depth.
        #In summary: Transpiling more aggressively leads to less qubits but more depth.#
        
        #Only letting mcx and logic synthesis survive has shown to be a good compromise
        
        transpiled_qc = transpile(qc, transpile_predicate = reordering_transpile_predicate)
        reordered_qc = reorder_qc(transpiled_qc)
        
        #Transpile logic synthesis
        reordered_qc = transpile(reordered_qc, transpile_predicate = lambda op : not (isinstance(op, PTControlledOperation) and op.base_operation.name in ["x"]))
        
        # We combine adjacent single qubit gates
        if not reordered_qc.abstract_params:
            reordered_qc = combine_single_qubit_gates(reordered_qc)
        
        #We now determine the amount of Qubits the circuit will need    
        required_qubits = 0
        max_required_qubits = 0
        
        for instr in reordered_qc.data:
            if instr.op.name == "qb_alloc":
                required_qubits += 1
            elif instr.op.name == "qb_dealloc":
                required_qubits -= 1
            if max_required_qubits < required_qubits:
                max_required_qubits += 1
        
        #Create a the QuantumCircuit. Note that we make sure that the Qubit naming is consistent,
        #since we don't want any non-deterministic elements in the function, as this can hinder bugfixing
        qc = QuantumCircuit()
        for i in range(workspace + max_required_qubits):
            qc.add_qubit(Qubit(str(i)))
    
        
        
        qc.clbits = list(qs.clbits)
    
    
        #This dictionary translates between the qubits of the input QuantumSession and the compiled circuit
        translation_dic = {}
        
        #This list contains the Qubits which are currently not allocated
        free_qb_list = list(qc.qubits)
        
        allocated_qb_list = []
        
        depth_dic = {b : 0 for b in qc.qubits + qc.clbits}
        
        
        #We now iterate through the data of the preprocessed QuantumCircuit
        for i in range(len(reordered_qc.data)):
            
            QuantumCircuit.fast_append = True
            instr = reordered_qc.data[i]
            if instr.op.name == "barrier":
                continue
            
            if instr.op.name == "qb_alloc":
                #We delay the allocation until an actual gate is executed
                allocated_qb_list.append(instr.qubits[0])
                continue
            
            #Check if any of the involved qubits need an allocation
            for qb in instr.qubits:
                
                if not qb in translation_dic:
                    #To pick a good allocation, we determine the depth dic and sort the
                    #available Qubits by their corresponding depth.
                    #Note that we add the hash value in order to prevent non-deterministic behavior
                    free_qb_list.sort(key = lambda x : depth_dic[x] + qc.qubits.index(x)*1E-5)
                    
                    translation_dic[qb] = free_qb_list.pop(0)
                    
                    qc.append(QubitAlloc(), [translation_dic[qb]])
    
            if instr.op.name == "qb_dealloc":
                #For the deallocation, we simply remove the qubits from the translation dic
                #and append it to the free_qb_list
                free_qb_list.append(translation_dic[instr.qubits[0]])
                allocated_qb_list.remove(instr.qubits[0])
                
                qc.append(instr.op, [translation_dic[qb] for qb in instr.qubits], instr.clbits)
                
                del translation_dic[instr.qubits[0]]
                
                continue
    
            if isinstance(instr.op, ControlledOperation) and instr.op.base_operation.name == "x" and instr.op.method == "auto" and len(instr.qubits) > 3:
                
                #This section deals with the automatic recompilation of mcx gates
                #The reason why we do this is, that it might not be clear how many free ancillae are
                #available for mcx gates. If we set the method to auto in the mcx function,
                #this section is called and replaces the generated mcx gate with an implementation
                #that is better fit to suit the amount of available ancillae.
                #We first determine the free ancillae
                
                
                clean_ancillae = list(free_qb_list)
                clean_ancillae.sort(key = lambda x : depth_dic[x] + qc.qubits.index(x)*1E-5)
                
                dirty_ancillae = list(set(translation_dic.values()) - set([translation_dic[qb] for qb in instr.qubits]))
                dirty_ancillae.sort(key = lambda x : depth_dic[x] + qc.qubits.index(x)*1E-5)
                
                QuantumCircuit.fast_append = False
                #This function generates the data for the hybrid implementation
                compiled_mcx_data = gen_hybrid_mcx_data(instr.qubits[:-1], instr.qubits[-1], instr.op.ctrl_state, clean_ancillae, dirty_ancillae)
                
                
                #We now append the data
                for qb in clean_ancillae + dirty_ancillae:
                    translation_dic[qb] = qb
                    
                for instr in compiled_mcx_data:
                    qc.append(instr.op, [translation_dic[qb] for qb in instr.qubits])
                    update_depth_dic(qc.data[-1], depth_dic)
                
                #And free up the qubits
                for qb in clean_ancillae + dirty_ancillae:
                    del translation_dic[qb]
                
                QuantumCircuit.fast_append = True
            
            #Finally if all of the above cases are not met, we simply append the operation to the translated qubits
            else:
                
                try:
                    qc.data.append(Instruction(instr.op, [translation_dic[qb] for qb in instr.qubits], instr.clbits))
                    
                except KeyError:
                    raise Exception("Found operation " + instr.op.name + " on unallocated qubit during compilation.")
                
                update_depth_dic(qc.data[-1], depth_dic)
        
        #Allocate the qubits which had an alloc gate but no actual gates
        for qb in allocated_qb_list:
            if qb not in translation_dic:
                translation_dic[qb] = free_qb_list.pop(0)
        
        
        #The following code is mostly about renaming and ordering the resulting circuit
        #in order to make the compiled circuit still comprehensible
        
        #We rename the allocated qubits to their name from the quantum session
        for i in range(len(qs.qubits)):
            if qs.qubits[i] in translation_dic:
                translation_dic[qs.qubits[i]].identifier = qs.qubits[i].identifier
                translation_dic[qs.qubits[i]].hash_value = qs.qubits[i].hash_value
    
        #We also want the qubits to be sorted according the order of the QuantumVariables
        #ie. the order of creation
        sorted_qubit_list = []
        for qv in qs.qv_list:
            
            #Due to the measurement reduction feature, not all qubits of quantum variables
            #that are live, are guaranteed to be represented in the translation dic
            #Therefore we need the try - except structure here
            try:
                sorted_qubit_list.extend([translation_dic[qb] for qb in qv.reg])
            except KeyError:
                pass
        
        #Furthermore, not all qubits that have been deallocated in the QuantumSession
        #are guaranteed to be removed from the translation dic, since the 
        #measurement_reduction function might have removed their uncomputation/deallocation gates
        temp = list(set(translation_dic.values()) - set(sorted_qubit_list))
        temp.sort(key = lambda x : x.identifier)
        sorted_qubit_list.extend(temp)
    
        
        #Finally, we rename the deallocated qubits to "workspace"    
        workspace_naming_counter = 0
        td_values = list(translation_dic.values())
        for i in range(len(qc.qubits)):
            if not qc.qubits[i] in td_values:
                qc.qubits[i].identifier = "workspace_" + str(workspace_naming_counter)
                workspace_naming_counter += 1
                sorted_qubit_list.append(qc.qubits[i])
    
        qc.qubits = sorted_qubit_list
        
        reduced_qc = reduce_depth(qc)
        
    if reduced_qc.depth() > qc.depth():
        return qc
    else:
        return reduced_qc


def gen_hybrid_mcx_data(controls, target, ctrl_state, clean_ancillae, dirty_ancillae):
    #This function generates the data for the hybrid mcx implementation
    
    from qrisp.circuit.multi_cx import hybrid_mcx
    from qrisp.core import QuantumVariable
    #Specify QuantumVariables to call mcx function
    control_qv = QuantumVariable(len(controls), name = "control")
    target_qv = QuantumVariable(1, name = "target")
    
    # dirty_ancillae = dirty_ancillae + clean_ancillae
    hybrid_mcx(control_qv, target_qv, ctrl_state = ctrl_state, num_ancilla = len(clean_ancillae), num_dirty_ancilla = len(dirty_ancillae))
    
    #Get the list of used ancillae    
    used_ancillae_set = set(control_qv.qs.qubits) - set(control_qv.reg) - set(target_qv.reg)
    
    #If we used the list() function to transform the set, this introduces a non-deterministic
    #element in the compilation algorithm, which can hamper bugfixing
    used_clean_ancillae = []
    used_dirty_ancillae = []
    
    for qb in control_qv.qs.qubits:
        if qb in used_ancillae_set:
            if "dirty" in qb.identifier:
                used_dirty_ancillae.append(qb)
            else:
                used_clean_ancillae.append(qb)
    
    
    depth_dic = get_depth_dic(control_qv.qs)
    
    used_clean_ancillae.sort(key = lambda x : -depth_dic[x])
    used_dirty_ancillae.sort(key = lambda x : -depth_dic[x])
    
    
    # ancilla_list = list(clean_ancillae)
    
    # for i in range(len(used_ancillae)):
    #     if "yong" in used_ancillae[i].identifier:
    #         used_ancillae.insert(0, used_ancillae.pop(i))
    #         ancilla_list.insert(0, dirty_ancillae[0])
    #         break
    
    #Now retarget the instructions such that they use the appropriate qubits
    data = control_qv.qs.data
    
    retarget_instructions(data, list(control_qv), controls)
    retarget_instructions(data, list(target_qv), [target])
    retarget_instructions(data, used_clean_ancillae, clean_ancillae)
    retarget_instructions(data, used_dirty_ancillae, dirty_ancillae)
    
    i = 0
    #Remove (de)allocation gates
    while i < len(data):
        if data[i].op.name in ["qb_dealloc", "qb_alloc"]:
            data.pop(i)
            continue
        i += 1
        
    return data



def reorder_qc(qc):
    
    from qrisp.uncomputation import dag_from_qc
    G = dag_from_qc(qc, remove_init_nodes = True)
    
    qc_new = qc.clearcopy()
    
    dealloc_identifier = lambda x : x.op.name == "qb_dealloc"
    alloc_identifer = lambda x : x.op.name == "qb_alloc"
    
    # mcx_identifier = lambda x : isinstance(x.op, PTControlledOperation) and x.op.base_operation.name == "x"
    # nmcx_identifier = lambda x : not mcx_identifier(x)
    # sub_sort = lambda G : topological_sort(G, prefer = mcx_identifier, delay = nmcx_identifier)
    # for n in topological_sort(G, prefer = dealloc_identifier, delay = alloc_identifer, sub_sort = sub_sort):
        
    for n in topological_sort(G, prefer = dealloc_identifier, delay = alloc_identifer):
        qc_new.append(n.instr)
    
    return qc_new



#This function performs a topological sort of the graph G where we try to execute any deallocation gates
#as early as possible while still adhering to the topological order.
#We to perform a depth-first search as described here: https://en.wikipedia.org/wiki/Topological_sorting
#According to the Wikipedia page, we are allowed to pick any node as a "starting point" of the DF-search,
#which allows us to have modify the algorithm such that it optimizes the (de)allocation order.
#The general idea is to pick the deallocation nodes as starting points, where we order them,
#such that those deallocation nodes that "require" the least allocation nodes are executed first.
#"Require" here means that there is a causal relationship between the allocation and deallocation nodes,
#ie. there is a path from the allocation node to the deallocation node.

#We can thus determine the amount of allocation nodes required for a deallocation node n by counting, the amount
#of allocation nodes in the "ancestors" subgraph of n.
def topological_sort(G, prefer = None, delay = None, sub_sort = nx.topological_sort):
    """
    Function to perform a topological sort on an Unqomp DAG which allows prefering/delaying specific types of nodes

    Parameters
    ----------
    G : nx.DiGraph
        The Unqomp DAG.
    prefer : function, optional
        Function which returns True, when presented with an Instruction, that should be prefered. The default is the function that returns False on all Operations
    delay : function, optional
        Function which returns True, when presented with an Instruction, that should be delayed. The default is the function that returns False on all Operations
    sub_sort : function, optional
        A function which performs a topological sort, which can sorting preferences of secondary importance. The default is nx.topological_sort.

    Returns
    -------
    lin : list[UnqompNode]
        The linearized list of UnqompNodes. The init nodes are not included.

    """
    
    if prefer is None:
        prefer = lambda x : False
    
    if delay is None:
        delay = lambda x : False
    
    G = G.copy()
    #Collect the prefered nodes
    prefered_nodes = []
    
    for n in G.nodes():
        if prefer(n.instr):
            
            prefered_nodes.append(n)
            
        n.processed = False
    
    #For large scales, finding the ancestors is a bottle neck. We therefore use a jitted version
    if len(G)*len(prefered_nodes) > 10000:
        anc_lists = ancestors(G, prefered_nodes)
        node_ancs = {prefered_nodes[i] : anc_lists[i] for i in range(len(prefered_nodes))}
    else:
        node_ancs = {prefered_nodes[i] : list(nx.ancestors(G, prefered_nodes[i])) for i in range(len(prefered_nodes))}
    #Determine the required delay nodes for each prefered nodes
    required_delay_nodes = {n : [] for n in prefered_nodes}
    
    for n in prefered_nodes:
        for k in node_ancs[n]:
            if delay(k.instr):
                required_delay_nodes[n].append(k)
    
    required_delay_nodes = {n : set(required_delay_nodes[n]) for n in prefered_nodes}
            
    #Generate linearization
    lin = []
    
    while prefered_nodes:
        
        #Sort nodes accordingly
        prefered_nodes.sort(key = lambda x : len(required_delay_nodes[x]))
        
        
        node = prefered_nodes.pop(0)
        ancs = []
        
        for n in node_ancs[node] + [node]:
            if n.processed:
                continue
            else:
                n.processed = True
                ancs.append(n)
        
        sub_graph = G.subgraph(ancs)
        
        lin += list(sub_sort(sub_graph))
        
        continue
        for n in prefered_nodes:
            required_delay_nodes[n] = required_delay_nodes[n] - required_delay_nodes[node]
        
    #Linearize the remainder
    remainder = []
    for n in G.nodes():
        if n.processed:
            continue
        else:
            n.processed = True
            remainder.append(n)
    
    # lin += list(sub_sort(G))
    lin += list(sub_sort(G.subgraph(remainder)))
    
    return lin


#Function to combine any sequences of single qubit gates into a single U3
def combine_single_qubit_gates(qc):
    
    def apply_combined_gates(qc_new, gate_list, qb):
        
        if not len(gate_list):
            return
        
        n = len(gate_list)
        
        m = np.eye(2)
        while gate_list:
            gate = gate_list.pop(-1)
            m = np.dot(m, gate.get_unitary())
        
        if np.linalg.norm(m - np.eye(2)) < 1E-10:
            return
            
        if n == 1:
            qc_new.append(gate, [qb])
            return
        
        qc_new.unitary(m, [qb])
        
    qb_dic = {qb : [] for qb in qc.qubits}
    
    qc_new = qc.clearcopy()
    
    for instr in qc.data:
        if len(instr.qubits) > 1 or instr.op.name in ["qb_alloc", "qb_dealloc"] or len(instr.clbits) > 0:
            
            for qb in instr.qubits:
                apply_combined_gates(qc_new, qb_dic[qb], qb)
            qc_new.append(instr)
        else:
            qb_dic[instr.qubits[0]].append(instr.op)
            
    for qb in qc.qubits:
        apply_combined_gates(qc_new, qb_dic[qb], qb)
    
    return qc_new

   

#This function reorders the circuit such that the intended measurements can be executed as early as possible.
#Additionally, any instructions that are not needed for the intended measurements are removed

#Intended measurements has to be a list of qubits

#The strategy is similar to the one presented in reorder_circuit:
#We bring the circuit in the dag representation and perform a topological sort with
#the intended measurement as prefered instructions

#After that we investigate the circuit for instructions that can be removed
def measurement_reduction(qc, intended_measurements):
    
    qc = qc.copy()
    
    #Insert intended measurements into the circuit
    for qb in intended_measurements:
        qc.measure(qb)

    #Generate dag representation    
    from qrisp.uncomputation import dag_from_qc
    G = dag_from_qc(qc, remove_init_nodes = True)
    
    #Create result qc
    qc_new = qc.clearcopy()

    #Define prefered instructions    
    measure_identifier = lambda x : x.op.name == "measure" and x.qubits[0] in intended_measurements
    
    
    #Perform topological sort
    for n in topological_sort(G, prefer = measure_identifier):
        qc_new.append(n.instr)

    #Check which instructions come after the final measurement        
    for i in range(len(qc_new.data))[::-1]:
        if measure_identifier(qc_new.data[i]):
            break
    
    redundant_qc = qc_new.clearcopy()
    
    redundant_qc.data = qc_new.data[i+1:]
    
    

    G = dag_from_qc(redundant_qc, remove_init_nodes = True)
    
    
    # alloc_graph = allocation_graph(redundant_qc)
    
    
    for node in G.nodes():
        
        if node.instr.op.name == "qb_dealloc":
            ancs = nx.ancestors(G, node)
            for pred in ancs:
                if pred.instr.op.name == "qb_alloc":
                    break
            else:
                redundant_qc.data.remove(node.instr)
                # print(f"removed {node.instr}")
                for pred in ancs:
                    
                    try:
                        redundant_qc.data.remove(pred.instr)
                    except ValueError:
                        pass
                    
                    
        
    # print(redundant_qc)    
    
    redundant_instructions = redundant_qc.data
    # print(redundant_instructions)
    
    
    # redundanct_instructions = qc_new.data[i+1:]
    # #Now we need to make sure we don't remove deallocation gates from the data
    # #because this would inflate the qubit count of the compiled circuit
    
    # #The strategy here is that if we find a deallocation gate
    # #we remove any instruction involving the deallocated qubit from the list 
    # #redundant instructions.
    
    # #If we however find an allocation gate, this chain of instructions can be
    # #safely removed and therefore stay in the redundant instructions
    
    # i = 0
    # while i < len(redundant_instructions):
    #     if redundant_instructions[i].op.name == "qb_dealloc":
    #         dealloc_qubit = redundant_instructions[i].qubits[0]
            
    #         j = 0
    #         while j < len(redundant_instructions):
    #             if dealloc_qubit in redundant_instructions[j].qubits:
    #                 if redundant_instructions[j].op.name == "qb_alloc":
    #                     i += 1
    #                     break
                    
    #                 redundant_instructions.pop(j)
    #                 continue
    #             j += 1
    #         continue
    #     i += 1
            
    #We now remove the redundant instructions and the inserted 
    #measurements from the circuit data
    i = 0
    while i < len(qc_new.data):
        
        if measure_identifier(qc_new.data[i]):
            qc_new.data.pop(i)
            continue
        if qc_new.data[i] in redundant_instructions:
            qc_new.data.pop(i)
            continue
        i += 1
    
    
    return qc_new


def allocation_graph(qc):
    from qrisp.uncomputation.unqomp import dag_from_qc
    dag = dag_from_qc(qc, remove_init_nodes = True)
    
    res = nx.DiGraph()
    
    dealloc_nodes = []
    for n in dag.nodes():
        if n.instr.op.name in ["qb_alloc", "qb_dealloc"]:
            res.add_node(n)
        if n.instr.op.name == "qb_dealloc":
            dealloc_nodes.append(n)
    
    for n in dealloc_nodes:
        for anc in nx.ancestors(n, dag):
            if anc.instr.op.name == "qb_alloc":
                res.add_edge(anc, n)
                
    return n
    
    
def update_depth_dic(instruction, depth_dic):
    
    if instruction.op.definition:
        qc = QuantumCircuit()
        qc.qubits = instruction.qubits
        qc.append(instruction)
        instr_list = qc.transpile().data
    else:
        instr_list = [instruction]
    
    # Assign each bit in the circuit a unique integer
    # to index into op_stack.
    # If no bits, return 0


    # Here we are playing a modified version of
    # Tetris where we stack gates, but multi-qubit
    # gates, or measurements have a block for each
    # qubit or cbit that are connected by a virtual
    # line so that they all stacked at the same depth.
    # Conditional gates act on all cbits in the register
    # they are conditioned on.
    # We treat barriers or snapshots different as
    # They are transpiler and simulator directives.
    # The max stack height is the circuit depth.
    
    for instr in instr_list:
        
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            continue
        qargs = instr.qubits
        cargs = instr.clbits
        
        levels = []
        # If count then add one to stack heights
        
        for b in qargs + cargs:
            # Add to the stacks of the qubits and
            # cbits used in the gate.
            levels.append(depth_dic[b] + 1)
    
        max_level = max(levels)
            
        for b in qargs + cargs:
            depth_dic[b] = max_level


@njit(cache = True)
def ancestors_jitted(start_index, indptr, indices, node_amount):
    
    to_do_array = np.zeros(node_amount, dtype = np.byte)
    to_do_array[start_index] = 1
    done_array = np.zeros(node_amount, dtype = np.byte)
    
    stack = 1
    while stack:
        
        node = np.argmax(to_do_array)
        to_do_array[node] = 0
        
        for i in range(indptr[node], indptr[node+1]):
            new_node = indices[i]
            if done_array[new_node] == 0:
                to_do_array[new_node] = 1
                stack += 1
        
        done_array[node] = 1
        stack -= 1
        
    return np.nonzero(done_array)[0]

@njit(parallel = True, cache = True)
def ancestors_jitted_wrapper(start_indices, indptr, indices, node_amount):
    
    res = [np.zeros(1, dtype = np.int64)]*len(start_indices)
    for i in prange(len(start_indices)):
        start_index = start_indices[i]
        res[i] = ancestors_jitted(start_index, indptr, indices, node_amount)
    
    return res
        

def ancestors(dag, start_nodes):
    
    node_list = list(dag.nodes())
    
    sprs_mat = nx.to_scipy_sparse_array(dag, format = "csc")
    
    start_indices = []
    for i in range(len(dag)):
        if node_list[i] in start_nodes:
            start_indices.append(i)
    
    res_list_indices = ancestors_jitted_wrapper(np.array(start_indices).astype(np.int32), sprs_mat.indptr, sprs_mat.indices.astype(np.int32), len(dag))
    
    node_list = [[node_list[j] for j in anc_indices] for anc_indices in res_list_indices]
    
    return node_list


# Function to cancel adjacent QFT, which are inverse to each other
# Due to the heavy use of Fourier arithmetic, this can happen alot
# especially if multiple arithmetic operation on a single target are executed
def qft_cancellation(qc):
    
    # The idea is to iterate through the instructions of the circuit
    # and save the last instruction acting on each qubit.
    # If a QFT is appended, we check if the last instruction was an inverse QFT
    # or an allocation and denote the index if so
    # Afterwards, we remove the QFTs or replace them with H gates
    
    last_instr_dic = {qb : None for qb in qc.qubits}
    cancellation_indices = []
    h_replacements = []
    dealloc_replacements = []
    from numpy.linalg import norm
    
    for i in range(len(qc.data)):
        
        instr = qc.data[i]
        if "QFT" in instr.op.name:
            previous_instruction_type = []
            
            for qb in instr.qubits:
                
                previous_instruction = qc.data[last_instr_dic[qb]]
                
                if previous_instruction.op.num_qubits != instr.op.num_qubits:
                    break
                
                if "QFT" in previous_instruction.op.name:
                    
                    if instr.op.num_qubits < 8:
                        
                        unitary_self = instr.op.get_unitary()
                        inv_unitary_other = previous_instruction.op.get_unitary().transpose().conjugate()

                        if bool(norm(unitary_self - inv_unitary_other) < 10**-4):
                            previous_instruction_type.append("QFT")
                        else:
                            break
                        
                    elif hash(instr.op.definition) == hash(previous_instruction.op.definition.inverse()):
                        previous_instruction_type.append("QFT")
                    else:
                        break
                    
                elif previous_instruction.op.name == "qb_alloc":
                   previous_instruction_type.append("alloc")
                else:
                    break
            else:
                if len(set(previous_instruction_type)) == 1:
                    if previous_instruction_type[0] == "QFT":
                        cancellation_indices.append(i)
                        cancellation_indices.append(last_instr_dic[qb])
                    else:
                        h_replacements.append(i)
        
        if instr.op.name == "qb_dealloc":
            
            deallocated_qubit = instr.qubits[0]
            previous_instruction = qc.data[last_instr_dic[deallocated_qubit]]
            
            if "QFT" in previous_instruction.op.name:
                try:
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                except AttributeError:
                    previous_instruction.deallocated_qubits = {qb : False for qb in previous_instruction.qubits}
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                    dealloc_replacements.append(last_instr_dic[deallocated_qubit])
                        
        for qb in instr.qubits:
            last_instr_dic[qb] = i
    
    
    for i in dealloc_replacements:
        instr = qc.data[i]
        for qb in instr.qubits:
            if not instr.deallocated_qubits[qb]:
                break
        else:
            h_replacements.append(i)
    
    new_qc = qc.clearcopy()
    
    # print(len(qc.data))
    for i in range(len(qc.data)):
        
        if i in h_replacements:
            for qb in qc.data[i].qubits:
                new_qc.h(qb)
            # print("H replacement successfull")
            continue
                
        if i in cancellation_indices:
            continue
        else:
            new_qc.append(qc.data[i])
    # print(len(new_qc.data))    
    # print("====")
    return new_qc
    


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


def array_entry_manipulation_test():
    
    from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, multi_measurement
    
    q_array = QuantumArray(QuantumBool(), (4,4))
    index_0 = QuantumFloat(2, signed = False)
    index_1 = QuantumFloat(2, signed = False)
    
    index_0[:] = 2
    index_1[:] = 1
    
    with q_array[(index_0, index_1)] as entry:
        
        entry.flip()
    
    res_array = q_array.most_likely()
    
    assert res_array[index_0.most_likely(), index_1.most_likely()]


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp import QuantumFloat, h

def test_measurement_reduction():
    
    qf = QuantumFloat(4)
    h(qf)
    qbls = []
    for i in range(100):
        qbls.append(qf == 2)

    assert qbls[50].get_measurement() == {False: 0.9375, True: 0.0625}


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 06.05.2022
from qrisp import QuantumFloat, QuantumArray, tensordot, dot
import numpy as np


def test_matrix_multiplication_example():
    qf = QuantumFloat(3,0, signed = False)

    test_1 = QuantumArray(qf, shape = (2,))
    test_2 = QuantumArray(qf, shape = (2,))


    test_1[:] = [2,3]
    test_2[:] = [1,1]
    # h(test_1)
    # h(test_2)
    print(test_1)
    assert (type(test_1[0]).__name__ and type(test_1[1]).__name__) == 'QuantumFloat' and type(test_1).__name__ == 'QuantumArray'
    assert str(test_1[0]) == '{2: 1.0}' and str(test_1[1]) == '{3: 1.0}'
    # test_1 = QuantumArray(qf, 3)
    # test_1.init_state([(np.array([1,2,3]), 1), (np.array([1,2,2]), 1j)])


    # test_1[:] = np.array([[5,6,2],[3,4,1],[1,1,0],[2,3,5]])

    # test_2 = QuantumArray(qf)
    # test_2[:] = np.eye(3)

    # res = test_1 @ test_2
    # np.sin(test_1)
    res = dot(test_1, test_2)
    print(res)
    assert type(res).__name__ == 'QuantumFloat'
    assert str(res) == '{5: 1.0}'
    # test = res.get_measurement(statevector = True, plot = True)



    qf = QuantumFloat(3,0, signed = False)

    test_1 = QuantumArray(qf, shape = (4,4))
    test_1.encode(np.eye(4))

    test_1 = test_1.reshape(2, 2, 2, 2)

    test_2 = QuantumArray(qf, shape = (2,2))
    test_2.encode(np.eye(2))


    res = tensordot(test_1, test_2, (-1, 0))


    qtype = QuantumFloat(3, signed = False)

    test_0 = np.round(np.random.rand(3,4)*5)
    test_1 = np.round(np.random.rand(4,2)*5)

    q_array = QuantumArray(qtype = qtype)


    q_array[:] = test_1


    test_0 @ q_array

    # test = semi_classic_matrix_multiplication(test_0, q_array)




"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import traceback

import numpy as np
import sympy

def bin_rep(n, bits):
    
    if n < 0:
        raise Exception("Only positive numbers are supported")
    
    if n >= 2**bits:
        raise Exception(str(n) + " can't be represented as a " + str(bits) + " bit number")
    
    zero_string = "".join(["0" for k in range(bits)])
    return (zero_string + bin(n)[2:])[-bits:]

def int_encoder(qv, encoding_number):
    if encoding_number > 2**len(qv) - 1:
        raise Exception("Not enough qubits to encode integer " + str(encoding_number))
    
    binary_rep = bin_rep(encoding_number, len(qv))[::-1]
    from qrisp import x
    for i in range(len(binary_rep)):
        if int(binary_rep[i]):
            
            x(qv[i])

#Calculates the binary expression of a given integer and returns it as an array of length bits
def int_as_array(k, bit):
    bin_str = bin_rep(k, bit)
    return np.array([int(c) for c in bin_str])

def array_as_int(array):
    result = 0
    for k in range(len(array)):
        if array[::-1][k]:
            result += 2**(k)
        
    return result

#Decomposes the circuit qc until no more decompositions are possible and then counts the cnot operations
def cnot_count(qc):
    
    qc = qc.transpile()
    
    gate_count_dic = qc.count_ops()
    
    try:
        return gate_count_dic["cx"]
    except KeyError:
        return 0



def is_inv(x, bit):
        
    #return (math.gcd(int(np.round(x, 3)),2**bit) == 1)
    
    #The only divisors 2**bit has is powers of 2
    #ie. if tha factorization of x doesn't contain any powers of 2 it is invertible
    #in other words: x is invertible if it is uneven
    return bool(int(x)%2)


def get_depth_dic(qc, transpile_qc = True):
    
    if len(qc.qubits) == 0:
        return {}
    
    
    if transpile_qc:
        qc = qc.transpile()
    
    # Assign each bit in the circuit a unique integer
    # to index into op_stack.
    bit_indices = {bit: idx for idx, bit in enumerate(qc.qubits + qc.clbits)}
    
    # If no bits, return 0
    if not bit_indices:
        return 0

    # A list that holds the height of each qubit
    # and classical bit.
    op_stack = [0] * len(bit_indices)

    # Here we are playing a modified version of
    # Tetris where we stack gates, but multi-qubit
    # gates, or measurements have a block for each
    # qubit or cbit that are connected by a virtual
    # line so that they all stacked at the same depth.
    # Conditional gates act on all cbits in the register
    # they are conditioned on.
    # We treat barriers or snapshots different as
    # They are transpiler and simulator directives.
    # The max stack height is the circuit depth.
    

    for instr in qc.data:
        
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            continue
        qargs = instr.qubits
        cargs = instr.clbits
        
        levels = []
        reg_ints = []
        # If count then add one to stack heights
        count = True
        
        for ind, reg in enumerate(qargs + cargs):
            # Add to the stacks of the qubits and
            # cbits used in the gate.
            reg_ints.append(bit_indices[reg])
            if count:
                levels.append(op_stack[reg_ints[ind]] + 1)
            else:
                levels.append(op_stack[reg_ints[ind]])
    
        max_level = max(levels)
        for ind in reg_ints:
            op_stack[ind] = max_level
    
    return {qc.qubits[i]: op_stack[i] for i in range(len(qc.qubits))}


def gate_wrap(*args, permeability = None, is_qfree = None, name = None, verify = False):
    """
    Decorator to bundle up the quantum instructions of a function into a single gate object.
    Bundled gate objects can help debugging as it allows for a more clear QuantumCircuit visualisation. 
    
    Furthermore bundling up functions is relevant for Qrisps uncomputation algorithm. 
    When bundling up for uncomputation, this decorator provides the means to annotate the gate objects with information about its permeability and qfree-ness. For further information about these concepts check the :ref:`uncomputation documentation<uncomputation>`. Specifying this information allows to skip the computationally costly automatic determination at runtime.
    
    Note that the specified information is not checked for correctness as this would defy the purpose.
    
    A shorthand for ``gate_wrap(permeability = "args", is_qfree = True)`` is the :meth:`lifted <qrisp.lifted>` decorator.
    
    
    .. warning::
        
        Using ``gate_wrap`` without specifying permeability and ``qfree``-ness on functions processing a lot of qubits,
        can causes long compile times, since the unitaries of these gates have to be determined numerically.
    
    .. warning::
        
        Incorrect information about permeability and ``qfree``-ness can yield incorrect compilation results. If you are unsure, use the ``verify`` keyword on a small scale first.
    
    
    Parameters
    ----------

    permeability : string or list, optional
        Specify the permeability behavior of the function. When given "args", it is assumed that the gate is permeable only on the qubits of the arguments. When given "full", it is assumed that the gate is permeable on every qubit it acts on (ie. also the result). When given a list of integers it is assumed, that the gate is permeable on the qubits of the arguments corresponding to the integers. The default is None.
    is_qfree : bool, optional
        Specify the qfree-ness of the function. The default is None.
    name : string, optional
        String which will be used for naming the gate object. The default is None.
    verify : bool, optional
        If set to ``True``, the specified information about permeability and ``qfree``-ness will be checked numerically. The default is ``False``.

    Examples
    --------

    We create a simple function wrapping up multiple gates: ::
        
        
        from qrisp import QuantumVariable, cx, x, h, z, gate_wrap
        
        @gate_wrap
        def example_function(a, b):
            
            cx(a,b)
            x(a)
            cx(b,a)
            h(b)
        
        a = QuantumVariable(3)
        b = QuantumVariable(3)
        
        example_function(a, b)
        
    >>> print(a.qs)
    QuantumCircuit:
    ---------------
         ┌───────────────────┐
    b.0: ┤0                  ├
         │                   │
    b.1: ┤1                  ├
         │                   │
    b.2: ┤2                  ├
         │  example_function │
    a.0: ┤3                  ├
         │                   │
    a.1: ┤4                  ├
         │                   │
    a.2: ┤5                  ├
         └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable a
    QuantumVariable b
    >>> print(a.qs.transpile())
         ┌───┐                    ┌───┐          
    b.0: ┤ X ├─────────────────■──┤ H ├──────────
         └─┬─┘┌───┐            │  └───┘┌───┐     
    b.1: ──┼──┤ X ├────────────┼────■──┤ H ├─────
           │  └─┬─┘┌───┐       │    │  └───┘┌───┐
    b.2: ──┼────┼──┤ X ├───────┼────┼────■──┤ H ├
           │    │  └─┬─┘┌───┐┌─┴─┐  │    │  └───┘
    a.0: ──■────┼────┼──┤ X ├┤ X ├──┼────┼───────
                │    │  ├───┤└───┘┌─┴─┐  │       
    a.1: ───────■────┼──┤ X ├─────┤ X ├──┼───────
                     │  ├───┤     └───┘┌─┴─┐     
    a.2: ────────────■──┤ X ├──────────┤ X ├─────
                        └───┘          └───┘     


    In the next example, we create a function that performs no quantum gates and specify that it is qfree and permeable on the second argument but not on the first. ::
        
        from qrisp import QuantumCircuit
        
        @gate_wrap(permeability = [1], is_qfree = True)
        def example_function(arg_0, arg_1):
            
            res = QuantumVariable(1)

            #Append an identity gate
            res.qs.append(QuantumCircuit(3).to_gate(), [arg_0, arg_1, res])
            
            return res
        
        qv_0 = QuantumVariable(1)
        qv_1 = QuantumVariable(1)
        
        res = example_function(qv_0, qv_1)
        
    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐
    qv_0.0: ┤0                  ├
            │                   │
    qv_1.0: ┤1 example_function ├
            │                   │
     res.0: ┤2                  ├
            └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv_0
    QuantumVariable qv_1
    QuantumVariable res     
    >>> qv_1.uncompute()
    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐
    qv_0.0: ┤0                  ├
            │                   │
    qv_1.0: ┤1 example_function ├
            │                   │
     res.0: ┤2                  ├
            └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv_0
    QuantumVariable res

    Since ``arg_1`` is marked as permeable, there are no further gates required for uncomputation.
    The situation is different for the other two QuantumVariables, where the qubits are not marked as permeable.
    
    >>> qv_0.uncompute(do_it = False)
    >>> res.uncompute()
    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐┌──────────────────────┐
    qv_0.0: ┤0                  ├┤0                     ├
            │                   ││                      │
    qv_1.0: ┤1 example_function ├┤1 example_function_dg ├
            │                   ││                      │
     res.0: ┤2                  ├┤2                     ├
            └───────────────────┘└──────────────────────┘
    Live QuantumVariables:
    ----------------------    
    """
    
    
    if len(args):
        return gate_wrap_inner(args[0])
    
    else:
        
        def gate_wrap_helper(function):
            return gate_wrap_inner(function, permeability = permeability, is_qfree = is_qfree, name = name, verify = verify)
        
        return gate_wrap_helper


def gate_wrap_inner(function, permeability = None, is_qfree = None, name = None, verify = False):
    
    
    def wrapped_function(*args, permeability = permeability, is_qfree = is_qfree, verify = verify, **kwargs):
        
        
        wrapped_function.__name__ = function.__name__
        from qrisp.core import recursive_qs_search, recursive_qv_search
        from qrisp.environments import GateWrapEnvironment 
        from qrisp.circuit import Qubit
        
        qs_list = recursive_qs_search([args, kwargs])

        if len(qs_list) == 0:
            raise Exception("Could not find QuantumSession object in function arguments during gate-wrapping")
        
        
        
        qs = qs_list[0]
        
        initial_qubits = set(qs.qubits)
        
        if name is None:
            gwe = GateWrapEnvironment(name = function.__name__)
        else:
            gwe = GateWrapEnvironment(name = name)
        
        with gwe:
            
            result = function(*args, **kwargs)
        
        try:
            qs.data.remove(gwe)
                
            gwe.compile()
                
                    
        except ValueError:
            pass
            
        if gwe.instruction is None:
            return result
        
        created_qubits = set(qs_list[0].qubits) - initial_qubits
        
        ancillas = []
        
        for qb in created_qubits:
            if qb.allocated == False:
                ancillas.append(qb)
        
        if not is_qfree is None:
            if verify and is_qfree:
                
                from qrisp.uncomputation import is_qfree as is_qfree_function
                
                if not is_qfree_function(gwe.instruction.op):
                    raise Exception(f"Verification of qfree-ness for function {function.__name__} failed")
            
            gwe.instruction.op.is_qfree = is_qfree
        
        if not permeability is None:
            permeability_dict = {}    
            arg_qubits = []
            
            
            if isinstance(permeability, list):
                args = [args[i] for i in permeability]
                permeability = "args"
                
            
            for arg in args:
                if isinstance(arg, list):
                    for item in arg:
                        if isinstance(item, Qubit):
                            arg_qubits.append(item)
                for arg_qv in recursive_qv_search([arg]):
                    arg_qubits += arg_qv.reg
            
            for i in range(len(gwe.instruction.qubits)):
                
                qb = gwe.instruction.qubits[i]
                
                if qb in arg_qubits:
                    
                    if permeability in ["args", "full"]:
                        permeability_dict[i] = True
                    else:
                        raise Exception("Don't know permeability option " + permeability)
                else:
                    if permeability == "full":
                        permeability_dict[i] = True
                    else:
                        permeability_dict[i] = False
        
            
            for i in range(len(gwe.instruction.qubits)):
                
                qb = gwe.instruction.qubits[i]
                
                if qb in ancillas:
                    #Even though ancilla qubits are permeable, we want to be able
                    #to use the gate_wrap decorator as an interface to perform recomputation.
                    #If we would mark them as permeable, Unqomp won't wrap the uncomputed
                    #gate in alloc/dealloc gates but instead wrap both the computation gate and the recomputation
                    #in alloc/dealloc gates
                    
                    #To undestand this behavior better consider the following example
                    
                        # from qrisp import QuantumFloat
                        # a = QuantumFloat(2)
                        # qf_res = a * a
                        # qf_res.uncompute()
                        # print(qf_res.qs)
                    
                    #If we mark the ancilla qubits as permeable, this gives
                    
                    # QuantumCircuit:
                    # ---------------
                    #              ┌──────────┐┌──────────┐┌─────────────┐              
                    #         a.0: ┤ qb_alloc ├┤0         ├┤0            ├──────────────
                    #              ├──────────┤│          ││             │              
                    #         a.1: ┤ qb_alloc ├┤1         ├┤1            ├──────────────
                    #              ├──────────┤│          ││             │┌────────────┐
                    #    return.0: ┤ qb_alloc ├┤2         ├┤2            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.1: ┤ qb_alloc ├┤3 __mul__ ├┤3 __mul___dg ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.2: ┤ qb_alloc ├┤4         ├┤4            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.3: ┤ qb_alloc ├┤5         ├┤5            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    # sbp_anc_0.0: ┤ qb_alloc ├┤6         ├┤6            ├┤ qb_dealloc ├
                    #              └──────────┘└──────────┘└─────────────┘└────────────┘
                    # Live QuantumVariables:
                    # ----------------------
                    # QuantumFloat a
                    
                    
                    # If we set them to non-permeable, we get instead
                    
                    # QuantumCircuit:
                    # ---------------
                    #              ┌──────────┐┌──────────┐                          ┌─────────────┐»
                    #         a.0: ┤ qb_alloc ├┤0         ├──────────────────────────┤0            ├»
                    #              ├──────────┤│          │                          │             │»
                    #         a.1: ┤ qb_alloc ├┤1         ├──────────────────────────┤1            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.0: ┤ qb_alloc ├┤2         ├──────────────────────────┤2            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.1: ┤ qb_alloc ├┤3 __mul__ ├──────────────────────────┤3 __mul___dg ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.2: ┤ qb_alloc ├┤4         ├──────────────────────────┤4            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.3: ┤ qb_alloc ├┤5         ├──────────────────────────┤5            ├»
                    #              ├──────────┤│          │┌────────────┐┌──────────┐│             │»
                    # sbp_anc_0.0: ┤ qb_alloc ├┤6         ├┤ qb_dealloc ├┤ qb_alloc ├┤6            ├»
                    #              └──────────┘└──────────┘└────────────┘└──────────┘└─────────────┘»
                    # «                           
                    # «        a.0: ──────────────
                    # «                           
                    # «        a.1: ──────────────
                    # «             ┌────────────┐
                    # «   return.0: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.1: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.2: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.3: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «sbp_anc_0.0: ┤ qb_dealloc ├
                    # «             └────────────┘
                    # Live QuantumVariables:
                    # ----------------------
                    # QuantumFloat a
                    
                    #In both cases, sbp_anc is recomputed, but in the second case, the reallocation allows the compiler
                    #to use the free qubit elsewhere and afterwards pick a potentially different qubit for the recomputation
                    
                    permeability_dict[i] = False
                
                for i in range(len(gwe.instruction.qubits)):
                    
                    if permeability_dict[i] is None:
                        permeability_dict[i] = False
                
                if verify:
                    from qrisp.uncomputation import is_permeable
                    permeable_qubit_indices = []
                    
                    for i in range(gwe.instruction.op.num_qubits):
                        if permeability_dict[i]:
                            permeable_qubit_indices.append(i)
                            
                    if not is_permeable(gwe.instruction.op, permeable_qubit_indices):
                        raise Exception(f"Verification of permeability for function {function.__name__} failed")
        
                
                gwe.instruction.op.permeability = permeability_dict
        
        return result
        
    
    
    return wrapped_function


#Function to measure multiple quantum variables at once to assess their entanglement
def multi_measurement(qv_list, shots = 10000, backend = None):
    """
    This functions facilitates the measurement of multiple QuantumVariables at the same time. This can be used if the entanglement structure between several QuantumVariables is of interest.

    Parameters
    ----------
    qv_list : list[QuantumVariable]
        A list of QuantumVariables.
    shots : int, optional
        The amount of shots to perform. The default is 10000.
    backend : BackendClient, optional
        The backend to evaluate the compiled QuantumCircuit on. By default, the backend from default_backend.py will be used.

    Raises
    ------
    Exception
        Tried to perform measurement with open environments.

    Returns
    -------
    counts_list : list
        A list of tuples. The first element of each tuple is a tuple again and contains the labels of the QuantumVariables. The second element is a float and indicates the probability of measurement.

    Examples
    --------
    
    We entangle three QuantumFloats via addition and perform a multi-measurement:
        
    >>> from qrisp import QuantumFloat, h, multi_measurement
    >>> qf_0 = QuantumFloat(4)
    >>> qf_1 = QuantumFloat(4)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 2
    >>> h(qf_1[0])
    >>> qf_sum = qf_0 + qf_1
    >>> multi_measurement([qf_0, qf_1, qf_sum])
    {(3, 2, 5): 0.5, (3, 3, 6): 0.5}  
    
    """
    
    
    if backend == None:
        if qv_list[0].qs.backend == None:
            from qrisp.default_backend import def_backend
            backend = def_backend
        else:
            backend = qv_list[0].qs.backend
        
    if len(qv_list[0].qs.env_stack) != 0:
        raise Exception("Tried to perform measurement with open environments")
    
    from qrisp import merge
    merge(qv_list)
    
    from qrisp.core.compilation import qompiler
    #Copy circuit in order to prevent modification
    from qrisp import QuantumArray, QuantumVariable, recursive_qv_search
    
    temp = recursive_qv_search(qv_list)
    
    compiled_qc = qompiler(qv_list[0].qs, intended_measurements = sum([qv.reg for qv in temp], []))
    # compiled_qc = qv_list[0].qs.copy()
    #Add classical registers for the measurement results to be stored in
    cl_reg_list = []
    
    
    for var in qv_list[::-1]:
        cl_reg = []
        
        if isinstance(var, QuantumArray):
            qubits = sum([qv.reg for qv in var.flatten()[::-1]], [])
        elif isinstance(var, QuantumVariable):
            qubits = var.reg
        else:
            raise Exception(f"Found type {type(var)} in measurement list")
        
        for i in range(len(qubits)):
            cl_reg.append(compiled_qc.add_clbit())
            
        cl_reg_list.append(cl_reg)
        
        #Add measurement instruction
        compiled_qc.measure(qubits, cl_reg)
    
    
    #counts = execute(qs_temp, backend, basis_gates = basis_gates, noise_model = noise_model, shots = shots).result().get_counts()
    counts = backend.run(compiled_qc, shots)
    counts = {k : counts[k] for k in sorted(counts)}
    
    #Convert the labeling bistrings of counts into list of labels
    new_counts = {}
    for i in range(len(counts)):
        
        #Retrieve the separated strings of each measurement variable
        
        counts_strings = []
        counts_bitstring = list(counts.keys())[i]
        
        bitstring_adress = 0
        for j in range(len(cl_reg_list)):
            cl_reg = cl_reg_list[::-1][j]
            counts_strings.append(counts_bitstring[bitstring_adress:bitstring_adress + len(cl_reg)][::-1])
            bitstring_adress += len(cl_reg)
        
        
        #Convert to integers and insert outcome labels
        counts_values = []
        for j in range(len(counts_strings)):
            
            outcome_int = int(counts_strings[j][::-1], 2)
            try:
                label = qv_list[j].decoder(outcome_int)
                if isinstance(label, np.ndarray):
                    from qrisp import OutcomeArray
                    label = OutcomeArray(label)
                counts_values.append(label)
            except AttributeError:
                counts_values.append(outcome_int)
        
        
        
        #Create array
        array_state = tuple(counts_values)
        try:
            new_counts[array_state] = counts[list(counts.keys())[i]]/shots
        except TypeError:
            raise Exception("Tried to create measurement outcome dic for QuantumVariable with unhashable labels")
        #Append to the counts list
        # counts_list.append((array_state, counts[list(counts.keys())[i]]/shots))
    
    #Sort counts_list such the most probable values come first
    new_counts = dict(sorted(new_counts.items(), key=lambda item: -item[1]))
    
    return new_counts



#Function to apply a phase function of signature phase_function(x,y,z..) -> float
#which specifies the phase for each constellation of outcome labels of the quantum variables in qv_list
def app_phase_function(qv_list, phase_function, t = 1, **kwargs):
    
    #Prepare the list of index tuples
    #For this we first create a list of outcome indices of each qv first
    index_lists = [list(range(2**qv.size)) for qv in qv_list]
    
    
    #We now calculate the direct product in order to obtain every possible combination
    from itertools import product
    product_index_list = list(product(*index_lists))
    
    #The next step is to iterate over every combination in order to determine the phases
    phases = []
    for i in range(len(product_index_list)):
        #Calculate the outcome labels of the current constellation of indices
        labels = [qv_list[j].decoder(product_index_list[i][j]) for j in range(len(qv_list))]
        
        #Calculate the phase
        phases.append(phase_function(*labels, **kwargs)*t)
    
    #Synthesize phase        
    from qrisp.logic_synthesis import gray_phase_synth_qb_list
    gray_phase_synth_qb_list(qv_list[0].qs, sum([qv.reg[::-1] for qv in qv_list], []), phases)


def as_hamiltonian(hamiltonian):
    r"""
    Decorator that recieves a regular Python function (returning a float) and returns a function of QuantumVariables, applying phases based on the function's output.
        
    Parameters
    ----------
    hamiltonian : function
        A function of arbitrary (non-quantum) variables returning a float.

    Returns
    -------
    hamiltonian_application : function
        A function of QuantumVariables, which applies the phase dictated by the hamiltonian to the corresponding states.
    
    Examples
    --------
    
        
    In this example we will demonstrate how a phase function with multiple arguments can be synthesized.
    For this we will create a phase function which encodes the fourier transform of different integers on the QuantumFloat x
    conditioned on the value of a QuantumChar ch.
    We will then apply the inverse Fourier transform to x and measure the results.
    ::
        
        import numpy as np
        from qrisp import QuantumChar, QuantumFloat, QFT, h
        #Create Variables
        x_size = 3
        x = QuantumFloat(x_size, 0, signed = False)
        
        ch = QuantumChar()
        
        #Bring x into uniform superposition so the phase function application yields 
        #a fourier transformed computation basis state
        h(x)
        
        
        #Bring ch into partial superposition (here |a> + |b> + |c> + |d>)
        h(ch[0])
        h(ch[1])
        
        from qrisp import multi_measurement, as_hamiltonian
        
        #In order to define the hamiltonian, we use regular Python syntax
        #The decorator "as_hamiltonian" turns it into a function 
        #that takes Quantum Variables as arguments. The decorator will add the 
        #keyword argument t to the function which mimics the t in exp(i*H*t)
        
        
        @as_hamiltonian
        def apply_multi_var_hamiltonian(ch, x):
            if ch == "a":
                k = 2
            elif ch == "b":
                k = 7
            elif ch == "c":
                k = 3
            else:
                k = 4
            
            #Return phase value
            #This is the phase distribution of the Fourier-transform
            #of the computational basis state |k>
            return k*x * 2*np.pi/2**x_size
        
        
        #Apply Hamiltonian
        apply_multi_var_hamiltonian(ch,x, t = 1)
        
        
        #Apply inverse Fourier transform
        QFT(x, inv = True)
        
        
    Acquire measurement results
    
    >>> multi_measurement([ch, x])
    {('a', 2): 0.25, ('b', 7): 0.25, ('c', 3): 0.25, ('d', 4): 0.25}
    
    We see that the measurement results correspond to what we specified in the Hamiltonian.
    
    In Bra-Ket notation, before applying the Hamiltonian, we are in the state
    
    .. math::
        
        \ket{\psi} = \frac{1}{\sqrt{4}}(\ket{a} + \ket{b} + \ket{c} + \ket{d}) \left( \frac{1}{\sqrt{8}} \sum_{x = 0}^8 \ket{x} \right)
    
    We then apply the Hamiltonian:
    
    .. math::
        
        \text{exp}(i\text{H(ch, x)})\ket{\psi} = \frac{1}{\sqrt{32}} ( &\ket{a} \sum_{x = 0}^{2^3-1} \text{exp}(2x \frac{2 \pi i}{2^3}) \ket{x} \\
                                                                     +&\ket{b} \sum_{x = 0}^{2^3-1} \text{exp}(7x \frac{2 \pi i}{2^3}) \ket{x} \\
                                                                    +&\ket{c} \sum_{x = 0}^{2^3-1} \text{exp}(3x \frac{2 \pi i}{2^3}) \ket{x} \\
                                                                    +&\ket{d} \sum_{x = 0}^{2^3-1} \text{exp}(4x \frac{2 \pi i}{2^3}) \ket{x})
    
    For each branch, the QuantumFloat tensor-factor is in a Fourier-transformed computational basis state. Thus, if we apply the inverse QFT, we receive:
        
    .. math::
        
        &\text{QFT}^{-1}\text{exp}(i\text{H(ch, x)})\ket{\psi} \\
            = & \frac{1}{\sqrt{4}} (\ket{a} \ket{2} + \ket{b} \ket{7} +\ket{c} \ket{3} + \ket{d} \ket{4})
                                                                    
    """
    
    
    def hamiltonian_application(*args, t = 1, **kwargs):
        gate_wrap(app_phase_function)(args, hamiltonian, t = t, **kwargs)
        # app_phase_function(args, hamiltonian, t = t, **kwargs)
    
    return hamiltonian_application

def perm_lock(qubits, message = ""):
    """
    Locks a list of qubits such that only permeable gates can be executed on these qubits. This means that an error will be raised if the user attempts to perform any operation involving these qubits if the operation does not commute with the Z-operator of this qubit. For more information, what a permeable gate is, check the  :ref:`uncomputation documentation <uncomputation>`. 
    
    This can be helpfull as it forbids all operations that change that computational basis state of this qubit but still allow controling on this qubit or applying phase gates.

    Using the keywoard message it is possible to extend the displayed error message.

    The effect of this function can be reversed using perm_unlock.

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The qubits to phase-tolerantly lock.
    message : str, optional
        The message why these qubits are locked.

    Examples
    --------
    
    We create a QuantumChar, perm-lock it's Qubits and attempt to initialize. 
    
    >>> from qrisp import QuantumChar, perm_lock, cx, p
    >>> q_ch_0 = QuantumChar()
    >>> perm_lock(q_ch_0)
    >>> q_ch_0[:] = "g"
    Exception: Tried to perform non-permeable operations on perm_locked qubits
    
    We now create a second QuantumChar and perform a CNOT gate
    
    >>> q_ch_1 = QuantumChar()
    >>> cx(q_ch_0[3], q_ch_1[2])
    
    Phase-gates are possible, too
    
    >>> p(0.1, q_ch_0)
    
    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list
    
    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                perm_lock(item)
                
            continue
        qb.perm_lock = True
        qb.perm_lock_message = message
        
def perm_unlock(qubits):
    """
    Reverses the effect of "perm_lock".

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The qubits to phase-tolerantly unlock.

    Examples
    --------
    
    We create a QuantumChar, perm-lock it's Qubits and attempt to initialize. 
    
    >>> from qrisp import QuantumChar, perm_lock, perm_unlock
    >>> q_ch = QuantumChar()
    >>> perm_lock(q_ch, message = "Qubits are perm-locked due to testing purposes")
    >>> q_ch[:] = "g"
    Exception: Qubits are perm-locked due to testing purposes
    
    >>> perm_unlock(q_ch)
    >>> q_ch[:] = "g"
    >>> print(q_ch)
    {'g': 1.0}
    
    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list
    
    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                perm_unlock(item)
            continue
        qb.perm_lock = False
        if hasattr(qb, "perm_lock_message"):
            del qb.perm_lock_message
        
def lock(qubits, message = ""):
    """
    Locks a list of qubits, implying an error will be raised if the user tries to perform any operation involving these qubits. 
    Using the keywoard message it is possible to extend the displayed error message.
    
    This can be reversed by calling unlock.

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The list of Qubits to lock.
    message : str, optional
        The message why these qubits are locked.

    Examples
    --------
    
    We create a QuantumChar, lock it's Qubits and attempt to initialize.
    
    >>> from qrisp import QuantumChar, lock
    >>> q_ch = QuantumChar()
    >>> lock(q_ch, message = "Qubits are locked due to testing purposes")
    >>> q_ch[:] = "g"
    Exception: Qubits are locked due to testing purposes
    
    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list
    
    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                lock(item)
            continue
        
        qb.lock = True
        qb.lock_message = message
        
def unlock(qubits):
    """
    Reverses the effect of "lock".

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The list of Qubits to lock.
    
    Examples
    --------
    
    We create a QuantumChar, lock it's Qubits and attempt to initialize.
    
    >>> from qrisp import QuantumChar, lock, unlock
    >>> q_ch = QuantumChar()
    >>> lock(q_ch)
    >>> q_ch[:] = "g"
    Exception: Tried to perform operations on locked qubits
    
    We now unlock and try again 
    
    >>> unlock(q_ch)
    >>> q_ch[:] = "g"
    >>> print(q_ch)
    {'g': 1.0}

    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list
    
    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                unlock(item)
            continue
        qb.lock = False
        if hasattr(qb, "lock_message"):
            del qb.lock_message
    
        

def benchmark_function(function):
    
    def benchmarked_function(*args, sort_stats = "tottime", stat_amount = 20, **kwargs):
        
        def slow_function():
            function(*args, **kwargs)
    
        import cProfile
        import pstats
        profile = cProfile.Profile()
    
        profile.runcall(slow_function)
    
        ps = pstats.Stats(profile)
        ps.strip_dirs()
        ps.sort_stats(sort_stats) 
        ps.print_stats(stat_amount)
    
    return benchmarked_function


def custom_qv(labels, decoder = None, qs = None, name = None):
    
    if not isinstance(labels, list):
        raise Exception("Tried to create custom QuantumVariable without providing a list type")
    
    if len(labels) == 0:
        raise Exception("Tried to create custom QuantumVariable without providing labels")
    elif len(labels) == 1:
        n = 1
    else:
        n = int(np.ceil(np.log2(len(labels))))
    
    from qrisp import QuantumVariable
    
    class CustomQuantumVariable(QuantumVariable):
        
        def __init__(self, qs = None, name = None):
            
            super().__init__(n, qs = qs, name = name)
        
        def decoder(self, x):
            
            
            
            if decoder == None:
                if x < len(labels):
                    return labels[x]
                else:
                    return "undefined_label_" + str(x)
                
            return decoder(x)
        
    return CustomQuantumVariable(qs = qs, name = name)


def init_state(qv, target_array):
    from qiskit.circuit.library.data_preparation.state_preparation import StatePreparation
    from qiskit import transpile
    
    qiskit_qc = StatePreparation(target_array).definition
    from qrisp import QuantumCircuit
    
    init_qc = QuantumCircuit.from_qiskit(qiskit_qc)
    
    #Find global phase correction
    from qrisp.simulator import statevector_sim
    
    init_qc.qubits.reverse()
    sim_array = statevector_sim(init_qc)
    init_qc.qubits.reverse()
    
    from numpy.linalg import norm
    
    
    arg_max = np.argmax(np.abs(sim_array))
    
    gphase_dif = (np.angle(target_array[arg_max]/sim_array[arg_max]))%(2*np.pi)
    
    init_qc.gphase(gphase_dif, 0)
    
    init_gate = init_qc.to_gate()
    
    init_gate.name = "state_init"
    
    qv.qs.append(init_gate, qv)
    
    

def get_statevector_function(qs, decimals = None):
    if len(qs.qv_list) == 0:
        return lambda x : 0
    else:
        
        from qrisp.simulator import statevector_sim
        
        
        compiled_qc = qs.compile()
        sv_array = statevector_sim(compiled_qc)
        
        if not decimals is None:
            sv_array = np.round(sv_array, decimals)
            
        def statevector(label_constellation, round = None):
            
            
            from qrisp import bin_rep
            
            qs = list(label_constellation.keys())[0].qs
            
            if len(label_constellation) != len(qs.qv_list):
                missing_variables = set([qv().name for qv in qs.qv_list]) - set([qv.name for qv in label_constellation.keys()])
                raise Exception("Tried to invoke statevector debugger without specifying an outcome label for each QuantumVariable registered in QuantumSession. Missing variables are: " + str(missing_variables))
            
            bitstring = len(compiled_qc.qubits)*["0"]
            
            for qf in label_constellation.keys():
                label_int = qf.encoder(label_constellation[qf])
                bin_label_int = bin_rep(label_int, qf.size)[::-1]
                
                for i in range(qf.size):
                    qubit_pos = compiled_qc.qubits.index(qf[i])
                    bitstring[qubit_pos] = bin_label_int[i]
            
            bitstring = "".join(bitstring)
            state_index = int(bitstring, base = 2)
            
            if round == None:
                return sv_array[state_index]
            else:
                return np.around(sv_array[state_index], round)
        
        return statevector
    


def check_if_fresh(qubits, qs):
    qc = qs.transpile()
    
    reversed_data = qc.data[::-1]
    for qb in qubits:
        for instr in reversed_data:
            if qb in instr.qubits:
                if instr.op.name == "qb_alloc":
                    break
                else:
                    return False
    
    return True

    

def get_measurement_from_qc(qc, qubits, backend, shots = 10000):
    
    #Add classical registers for the measurement results to be stored in
    cl = []
    for i in range(len(qubits)):
        cl.append(qc.add_clbit())
    
    #Add measurement instruction
    for i in range(len(qubits)):
        qc.measure(qubits[i], cl[i])
    
    #Execute circuit
    counts = backend.run(qc, shots)
    
    #Remove other measurements outcomes from counts dic
    new_counts_dic = {}
    for key in counts.keys():
        #Remove possible whitespaces
        new_key = key.replace(" ", "")
        #Remove other measurements
        new_key = new_key[-len(cl):]
        
        new_key = int(new_key, base = 2)
        try:
            new_counts_dic[new_key] += counts[key]
        except KeyError:
            new_counts_dic[new_key] = counts[key]
    
    
    counts = new_counts_dic
    #Plot result (if needed)
    
    #Normalize counts
    for key in counts.keys():
        counts[key] = counts[key]/shots
    
    return counts



def find_calling_line(level = 0):
    stack = traceback.extract_stack(limit = level+3)
    return str(traceback.format_list(stack)[1].split('\n')[1].strip())  # prints "a = fct1()"


def retarget_instructions(data, source_qubits, target_qubits):
    from qrisp import QuantumEnvironment
    
    for i in range(len(data)):
        
        instr = data[i]
        
        if isinstance(instr, QuantumEnvironment):
            retarget_instructions(instr.original_data, source_qubits, target_qubits)
            retarget_instructions(instr.env_data, source_qubits, target_qubits)
            continue
        
        for j in range(len(instr.qubits)):
            if instr.qubits[j] in source_qubits:
                instr.qubits[j] = target_qubits[source_qubits.index(instr.qubits[j])]


def redirect_qfunction(function_to_redirect):
    """
    Decorator to turn a function returning a QuantumVariable into an in-place function. This can be helpful for manual uncomputation if we have a function returning some QuantumVariable but we want the result to operate on some other variable, which is supposed to be uncomputed.
    
    Parameters
    ----------
    function_to_redirect : function
        A function returning a QuantumVariable.

    Raises
    ------
    Exception
        Given function did not return a QuantumVariable
    Exception
        Tried to redirect quantum function into QuantumVariable of differing size

    Returns
    -------
    redirected_function : function
        A function which performs the same operation as the input but now has the keyword argument target. Every instruction that would have been executed on the input functions result is executed on the QuantumVariable specified by target instead.
        
    
    Examples
    --------
    
    We create a function that determins the AND value of it's inputs and redirect it onto another QuantumBool. ::
        
        from qrisp import QuantumBool, mcx, redirect_qfunction
        
        #This function has only two arguments and returns it's result
        def AND(a, b):
            
            res = QuantumBool()
            
            mcx([a,b], res)
            
            return res
        
        a = QuantumBool(name = "a")
        b = QuantumBool(name = "b")
        c = QuantumBool(name = "c")
        
        #This function has two arguments and the keyword argument target
        redirected_AND = redirect_qfunction(AND)
        
        redirected_AND(a, b, target = c)
        
        
    >>> print(a.qs)
    QuantumCircuit:
    ---------------
    b.0: ──■──
           │  
    a.0: ──■──
         ┌─┴─┐
    c.0: ┤ X ├
         └───┘
    Live QuantumVariables:
    ----------------------
    QuantumBool b
    QuantumBool a
    QuantumBool c
        
        

    """
    from qrisp import QuantumEnvironment, merge, QuantumVariable
    
    def redirected_qfunction(*args, target = None, **kwargs):
        
        
        env = QuantumEnvironment()
        env.manual_allocation_management = True
        
        with env:
            res = function_to_redirect(*args, **kwargs)
            
            if not isinstance(res, QuantumVariable):
                raise Exception("Given function did not return a QuantumVariable")
            
            
            target = list(target)
            
            
            if len(res) != len(target):
                raise Exception("Tried to redirect quantum function into QuantumVariable of differing size")
                
            i = 0
            res_is_new = False
            while i < len(env.env_qs.data):
                
                instr = env.env_qs.data[i]
                if isinstance(instr, QuantumEnvironment):
                    pass
                elif instr.op.name == "qb_alloc" and instr.qubits[0] in list(res):
                    env.env_qs.data.pop(i)
                    res_is_new = True
                    continue
                
                i += 1
            
            retarget_instructions(env.env_qs.data, list(res), target)
        
        if res_is_new:
            #Remove all traces of res
            res.delete()

            for i in range(res.size):
                res.qs.qubits.remove(res[i])
                res.qs.data.pop(-1)
            
            for i in range(len(res.qs.deleted_qv_list)):
                qv = res.qs.deleted_qv_list[i]
                if qv.name == res.name:
                    res.qs.deleted_qv_list.pop(i)
                    break
                
        return target
    
    redirected_qfunction.__name__ = function_to_redirect.__name__
    
    return redirected_qfunction


def get_sympy_state(qs, decimals):
    
    from sympy.physics.quantum import Ket, OrthogonalKet
    from sympy import nsimplify, simplify, pi, count_ops, exp, I, cancel, factor, simplify, cos, Rational, Symbol, sin

    from qrisp.simulator import statevector_sim
    
    qv_list = list(qs.qv_list)
    
    labels = []
    for qv in qv_list:
        labels.append([qv.decoder(i) for i in range(2**qv.size)])
        
    compiled_qc = qs.compile()
    
    sv_array = statevector_sim(compiled_qc)
    
    if not sv_array.dtype == np.dtype("O"):
        angles = np.angle(sv_array)%(2*np.pi)/(np.pi)
        
        if not decimals is None:
            sv_array = np.round(sv_array, decimals)
            angles = np.round(angles, decimals)
    
        nz_indices = np.nonzero(sv_array)[0]
        nnz = len(nz_indices)
    
    else:
        import sympy as sp
        
        nz_indices = []
        
        for i in range(len(sv_array)):
            
            entry = simplify(sv_array[i])
            
            for a in sp.preorder_traversal(entry):
                if isinstance(a, sp.Float):
                    entry = entry.subs(a, round(a, decimals))
            
            sv_array[i] = entry
            
            if not sv_array[i] == 0:
                nz_indices.append(i)
            
        nnz = len(nz_indices)
    
    res = 0
    for ind in list(nz_indices):
        amplitude = sv_array[ind]
        
        if not sv_array.dtype == np.dtype("O"):
            
            abs_amp = trigify_amp(amplitude, decimals, nnz)
            
            #For some reason there is a sympy error, when the angle is equal to 1        
            if angles[ind] == 1:
                phase = 1
            else:
                phase = nsimplify(angles[ind], tolerance = 10**(decimals-1))
            
            if count_ops(phase) > 5:
                phase = angles[ind]
            
            ket_expr = exp(I*phase*pi)*abs_amp*nnz**0.5
            
        else:
            
            process_stack = [amplitude]
            while process_stack:
                
                a = process_stack.pop(0)
                if isinstance(a, (sympy.core.add.Add, sympy.core.mul.Mul)) and len(a.free_symbols) != 0:
                    process_stack.extend(a.args)
                
                elif len(a.free_symbols) == 0:
                    
                    sub_float = np.round(complex(a.evalf()), decimals)
                    
                    if np.abs(sub_float - 1) < 10**-decimals:
                        abs_amp = 1
                        continue
                    elif np.abs(sub_float) < 10**-decimals:
                        entry = entry.subs(a, 0)
                        continue
                    elif np.abs(sub_float) > 1:
                        continue
                    else:
                        abs_amp = trigify_amp(sub_float, decimals, nnz)
                    
                    if np.angle(complex(a.evalf()))/np.pi == 1:
                        phase = -1
                    else:
                        phase = sp.exp(sp.I*nsimplify(np.angle(complex(a.evalf()))/np.pi, tolerance = 10**(decimals-1))*Symbol("pi"))
                    
                    
                    expr = abs_amp*phase
                    
                    amplitude = amplitude.subs(a, expr)
            
            amplitude = amplitude.subs(1j, sp.I)
            
            ket_expr = sp.trigsimp(amplitude)*nnz**0.5

            
        int_string = bin_rep(ind, len(compiled_qc.qubits))
        
        labels = []
        for qv in qv_list:
            bit_string = ""
            for qb in qv.reg:
                bit_string += int_string[compiled_qc.qubits.index(qb)]
            
            label = qv.decoder(int(bit_string[::-1], 2))
            ket_expr *= OrthogonalKet(str(label))
            
        res += ket_expr


    res = cancel(nsimplify(1/nnz**0.5) * res)
    
    if isinstance(res, sympy.core.mul.Mul):
        temp = 1
        for arg in res.args[:-1]:
            temp *= nsimplify(arg.subs({Symbol("pi") : pi}))
        
        res = temp * res.args[-1]
    
    res = res.subs({Symbol("pi") : pi})
    return res
            

def trigify_amp(amplitude, decimals, nnz):
    from sympy import nsimplify, simplify, pi, count_ops, exp, I, cancel, factor, simplify, cos, Rational, Symbol, sin, latex
    
    cos_expr = nsimplify(np.arccos(np.abs(amplitude))/np.pi, tolerance = 10**-(decimals-1))
    sin_expr = nsimplify(np.arcsin(np.abs(amplitude))/np.pi, tolerance = 10**-(decimals-1))

    # if count_ops(sin_expr) > count_ops(cos_expr):
    if len(latex(sin_expr)) > len(latex(cos_expr)):
        expr = "cos"
        temp = cos_expr
    # elif count_ops(sin_expr) < count_ops(cos_expr):
    elif len(latex(sin_expr)) < len(latex(cos_expr)):
        expr = "sin"
        temp = sin_expr
    elif len(sin_expr.free_symbols) == 0:
        if sin_expr.evalf() > cos_expr.evalf():
            expr = "cos"
            temp = cos_expr
        else:
            expr = "sin"
            temp = sin_expr
    else:
        temp = nsimplify(np.abs(amplitude)*nnz**0.5, tolerance = 10**-(decimals-1))/nnz**0.5
    
    # if count_ops(temp) > 4:
    if len(latex(temp)) > 20:
        temp = nsimplify(np.abs(amplitude)*nnz**0.5, tolerance = 10**-(decimals-1))/nnz**0.5
        if len(latex(temp)) > 20:
            abs = np.abs(amplitude)
            
        else:
            abs = temp
    
    else:
        if expr == "cos":
            abs = cos(cos_expr*Symbol("pi"))
        else:
            abs = sin(sin_expr*Symbol("pi"))
    
    return abs


def render_qc(qc):
    latex_str = qc.to_latex()
    import tempfile
    import os.path
    import subprocess
    from IPython.display import Image, display
    
    with tempfile.TemporaryDirectory(prefix="texinpy_") as tmpdir:
        path = os.path.join(tmpdir, "document.tex")
        with open(path, 'w') as fp:
            fp.write(latex_str)
        subprocess.run(["lualatex", path], cwd=tmpdir)
        subprocess.run(["pdftocairo", "-singlefile", "-transp", "-r", "100", "-png", "document.pdf", "document"], cwd=tmpdir)

        im = Image(filename=os.path.join(tmpdir, "document.png"))
        display(im)

def lifted(*args, verify = False):
    """
    Shorthand for ``gate_wrap(permability = "args", is_qfree = True)``.
    
    A lifted function is ``qfree`` and permeable on it's inputs. The results of lifted functions can be automatically uncomputed even if they contain functions that could not be uncomputed on their own.
    
    You can find more information about these concepts :ref:`here <Uncomputation>` or `here <https://silq.ethz.ch/overview#/overview/3_uncomputation>`_. Note that the concept of permeability in Qrisp is a more general version of Silq's ``const``.
    
    .. warning::
        
        Incorrect information about permeability and ``qfree``-ness can yield incorrect compilation results. If you are unsure, use the ``verify`` keyword on a small scale first.


    Parameters
    ----------
    
    verify : bool, optional
        If set to ``True``, the specified information about permeability and ``qfree``-ness will be checked numerically. The default is ``False``.

    Examples
    --------
    
    We create a function performing the `Margolus gate <https://arxiv.org/abs/quant-ph/0312225>`_. As it contains ``ry`` rotations, there are non-``qfree`` steps involved.
    Putting on the ``lifted`` decorator however marks the function as ``qfree`` as a whole.
    
    ::
        
        from qrisp import QuantumVariable, cx, ry, lifted
        from numpy import pi
        
        @lifted(verify = True)
        def margolus(control):
            
            res = QuantumVariable(1)
            ry(pi/4, res)
            cx(control[1], res)
            ry(-pi/4, res)
            cx(control[0], res)
            ry(pi/4, res)
            cx(control[1], res)
            ry(-pi/4, res)
            
            return res
        
        
        control = QuantumVariable(2)
        res = margolus(control)
        
    >>> print(res.qs)
    QuantumCircuit:
    ---------------
               ┌───────────┐
    control.0: ┤0          ├
               │           │
    control.1: ┤1 margolus ├
               │           │
        res.0: ┤2          ├
               └───────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumVariable res
    
    >>> res.uncompute()
    >>> print(res.qs)
    QuantumCircuit:
    ---------------
               ┌───────────┐┌──────────────┐
    control.0: ┤0          ├┤0             ├
               │           ││              │
    control.1: ┤1 margolus ├┤1 margolus_dg ├
               │           ││              │
        res.0: ┤2          ├┤2             ├
               └───────────┘└──────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    
    Note that we set the ``verify`` keyword to ``True`` in this example. In more complex functions, involving many qubits this feature should only be used for bug-fixing on a small scale, since the verification can be time consuming.

    """
    
    if len(args) == 0:
        
        def lifted_helper(function):
            
            return gate_wrap(permeability = "args", is_qfree = True, verify = verify)(function)
        
        return lifted_helper
    
    else:
        return gate_wrap(permeability = "args", is_qfree = True)(args[0])
        
            
    



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.arithmetic import QuantumFloat
from qrisp.iterators.qrange import qRange
from qrisp import h, merge

def test_loop():
    from qrisp import QuantumFloat, qRange, h
    # Create some QuantumFloats
    n = QuantumFloat(3)
    qf = QuantumFloat(5)
    # Initialize the value 4
    n[:] = 4


    # Apply H-gate to 0-th qubit
    h(n[0])
    
    assert n.get_measurement() == {4.0: 0.5, 5.0: 0.5}

    # Perform successive addition of increasing numbers
    for i in qRange(n):
        qf += i

    assert qf.get_measurement() == {10.0: 0.5, 15.0: 0.5}




"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import threading

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol
from thrift.server import TServer, TNonblockingServer 
from qrisp.interface.thrift_interface.stoppable_thrift_server import StoppableThriftServer

from qrisp.quantum_network.interface.codegen.ttypes import Message

from qrisp.interface.circuit_converter import convert_circuit

from qrisp.quantum_network.qn_simulator_server import QuantumNetwork

from qrisp.interface import PortableQubit
from qrisp import Qubit

# Returns the hosts ip
def get_ip():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP



class QuantumNetworkHandler(QuantumNetwork):
    
    def run(self, qc, name):
        converted_circuit = convert_circuit(qc, "qrisp")
        return super().run(converted_circuit, name)

    def get_clear_qc(self, name):
        
        qc = super().get_clear_qc(name)
        
        return convert_circuit(qc, "thrift")
    
    
    def request_qubits(self, amount, name):
        
        qb_list = super().request_qubits(amount, name)
        
        return [PortableQubit(qb.identifier) for qb in qb_list]


    def send_qubits(self, sender, recipient, msg):
        
        super().send_qubits(sender, recipient, [Qubit(qb.identifier) for qb in msg.qubits], msg.annotation)


    def get_overall_qc(self):
        
        qc = super().get_overall_qc()
        
        return convert_circuit(qc, "thrift")
    
    def inbox(self, name):
        
        inbox_list = super().inbox(name)
        
        res_list = []
        
        for i in range(len(inbox_list)):
            
            res_list.append(Message([PortableQubit(qb.identifier) for qb in inbox_list[i][0]], inbox_list[i][1]))

        return res_list

from qrisp.quantum_network.interface.codegen import QuantumNetworkService
 
class QuantumNetworkServer:
    """
    This class sets up a server to coordinate the QuantumNetwork. All the simulations are performed in this instance, so the machine running it should preferably have abundant computational resources.
    
    Parameters
    ----------
    socket_ip : str, optional
        The IP address of the network socket. By default, the IP adress of the executing machine is used.
    port : int, optional
        The port to listen for requests. The default is 7070.
    background : bool, optional
        If set to True, the server will run in the background as a separate thread. The default is False.
    
    Examples
    --------
    
    We create a server listening to the localhost IP address:
    
    >>> from qrisp.quantum_network import QuantumNetworkServer
    >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1")
    >>> example_server.start()
    
    """
    
    def __init__(self, socket_ip = None, port = None, background = False):
        
        if socket_ip == None:
            socket_ip = get_ip()
        
        self.socket_ip = socket_ip
        
        self.background = background
        self.port = port
        
        
    #Starts the server
    def start(self):
        """
        Starts the server.

        Returns
        -------
        None.

        """
        
        
        if self.port == None:
            self.port = 7070
        
        
        self.thread_name = "qn" + ":" + str(self.port)
        
        self.handler = QuantumNetworkHandler()
        
        
        processor = QuantumNetworkService.Processor(self.handler)
        tfactory = TTransport.TBufferedTransportFactory()

        self.transport = TSocket.TServerSocket(host=self.socket_ip, port=self.port)    
            
        pfactory = TBinaryProtocol.TBinaryProtocolFactory()
        
        #server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)
        # self.server = TServer.TThreadedServer(processor, self.transport, tfactory, pfactory, daemon = True)
        self.server = StoppableThriftServer(processor, self.transport, tfactory, pfactory, daemon = True)
        
        # self.server = TNonblockingServer.TNonblockingServer(processor, self.transport, tfactory, pfactory)
        
        if self.background:        
            #Create thread
            self.thr = threading.Thread(target = self.server.serve, name = self.thread_name)
            self.thr.setDaemon(True)
            
            #Start the thread
            self.thr.start()
        else:
            self.server.serve()
        
    
    def stop(self):
        """
        Stops the server.

        Returns
        -------
        None.

        """
        if hasattr(self, "server"):
            self.server.stop()
    
    def __del__(self):
        self.handler.reset_network_state()
        self.stop()


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 22.07.2022
from qrisp import QuantumString, QuantumChar

def test_string_test():
    q_str = QuantumString()

    q_str_2 = QuantumString()

    q_ch = QuantumChar()

    q_str[:] = "hello"

    q_ch[:] = " "

    q_str_2[:] = "world"


    q_str += q_ch


    q_str += q_str_2


    q_str += "! "


    q_str_3 = q_str.duplicate(init = True)

    assert (type(q_str).__name__ and type(q_str_3).__name__) == 'QuantumString'
    assert str(q_str + q_str_3) == "{'hello world! hello world! ': 1.0}"
    print(q_str + q_str_3)


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


def test_mcx():
    
    from qrisp import h, mcx, QuantumVariable, multi_measurement, bin_rep
    import numpy as np
    def test_mcx_inner(n, ctrl_state, method, test_phase = True, num_ancilla = 0):
    
        ctrl = QuantumVariable(n)
        
        target = QuantumVariable(1)
        
        h(ctrl)
        
        if method in ["hybrid", "balauca_dirty"]:
            mcx(ctrl, target, method = method, ctrl_state = ctrl_state, num_ancilla = num_ancilla)
        else:
            mcx(ctrl, target, method = method, ctrl_state = ctrl_state)
        mes_res = multi_measurement([ctrl, target])
        statevector = ctrl.qs.statevector("array")
        
        #Test correct flipping behavior
        for i in range(2**n):
            ctrl_var = bin_rep(i, n)
            if ctrl_var != ctrl_state:
                assert (ctrl_var, "0") in mes_res
            else:
                assert (ctrl_var, "1") in mes_res
        
        if not test_phase:
            return
        
        angles = np.angle(statevector[np.abs(statevector) > 1/2**(n/2+1)])
        
        #Test correct phase behavior
        assert np.sum(np.abs(angles)) < 0.1
        
    for n in range(1, 8):
        print("n: ", n)
        for ctrl_state in [n*"1", n*"0", bin_rep(int(np.random.randint(0, 2**n-1)), n)]:
            print("Ctrl state: ", ctrl_state)
            for method in ["gray", "gray_pt", "maslov", "balauca", "yong"]:
                print(method)
                
                if method == "maslov" and n > 4:
                    continue
            
                test_mcx_inner(n, ctrl_state, method, test_phase = (method != "gray_pt"))
                
            for method in ["hybrid", "balauca_dirty"]:
                for num_ancilla in range(8):
                    print("Ancilla count: ", num_ancilla)
                    test_mcx_inner(n, ctrl_state, method, num_ancilla = num_ancilla)
                
            
    




"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 05.05.2022
from qrisp import QuantumVariable, x, h, merge
from qrisp.environments.GMS_environment import GMSEnvironment
import numpy as np


def test_GMS_environment_example():
    qv1 = QuantumVariable(1)
    qv2 = QuantumVariable(1)
    
    merge([qv1, qv2])
    
    
    qs = qv1.qs
    test = GMSEnvironment()
    
    
    h(qv1)
    x(qv2)
    with test:
        
        qs.cp(np.pi/2, 0,1)
        qs.p(np.pi/2, 0)
        qs.p(np.pi/2, 1)
        
    
    h(qv1)

    ###################
    from qrisp.arithmetic import QuantumFloat

    
    qf = QuantumFloat(3)

    qf.encode(1)

    from qrisp import QFT

    QFT(qf, use_gms=True)

    sv = qf.qs.statevector("array")



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""



from qrisp.simulator import ImpureQuantumState, run, QuantumState, single_shot_sim        
from qrisp import Qubit, Clbit, QuantumCircuit

class QuantumNetwork:
    
    def __init__(self):
        
        self.qs = QuantumState(0)
        self.qubit_map = {}
        self.overall_circuit = QuantumCircuit()
        self.client_measurement_counter = {}
        self.inbox_dict = {}
        self.qubit_creation_counter = {}
        
    def register_client(self, name):
        if name in list(self.client_measurement_counter.keys()):
            raise Exception("Tried to register an existing client.")
            
        self.client_measurement_counter[name] = 0
        self.inbox_dict[name] = []
        self.qubit_creation_counter[name] = 0
    
    def send_qubits(self, sender, recipient, qubits, annotation):
        
        keys = list(self.qubit_map.keys())
        ids = keys
        
        qubits = convert_qubit_names(qubits, sender)
        
        for qb in qubits:
            
            if qb.identifier in ids:
                index = ids.index(qb.identifier)
            
                if sender != self.qubit_map[keys[index]]:
                    raise Exception("Tried to send qubit not belonging to sender.")
            else:
                self.add_qubit(qb, sender)
                
            self.qubit_map[qb.identifier] = recipient
            
        self.inbox_dict[recipient].append((qubits, annotation))
        

    def get_clear_qc(self, name):
        
        qc = QuantumCircuit()
        
        for key in self.qubit_map.keys():
            if self.qubit_map[key] == name:
                qc.add_qubit(Qubit(key))
        
        return qc
    
    def request_qubits(self, amount, owner):
        
        res_list = []        
        for i in range(amount):
            new_qb = Qubit("qb_" + str(self.qubit_creation_counter[owner]) + "@" + owner)
            self.add_qubit(new_qb, owner)
            res_list.append(new_qb)
        
        return res_list
            
    def add_qubit(self, qubit, owner):
        
        if not owner in self.client_measurement_counter.keys():
            raise Exception("Tried to request qubits for non registered client")
        
        qubit = convert_qubit_names([qubit], owner)[0]
        
        if qubit in self.qubit_map.keys():
            raise Exception("Tried to create already existing qubit")
        
        self.qubit_map[qubit.identifier] = owner
        self.qubit_creation_counter[owner] += 1
        self.overall_circuit.add_qubit(qubit)
        self.qs.add_qubit()
        
        return [qubit]
    
        
                
    def run(self, qc, name):
        
        keys = list(self.qubit_map.keys())
        ids = keys
        
        qc = convert_qc_qubit_names(qc, name)
        
        for qb in qc.qubits:
            
            if qb.identifier in ids:
                index = ids.index(qb.identifier)
                
                if self.qubit_map[keys[index]] != name:
                    raise Exception("Tried to operate one foreign qubits")
            
            else:
                self.add_qubit(qb, name)
                
            
        
        temp_qc = self.overall_circuit.clearcopy()
        temp_qc.clbits = []
        
        translation_dic = {}
        
        for clbit in qc.clbits:
            overall_qc_clbit = Clbit("cb" + "_" + name + "_" + str(self.client_measurement_counter[name]))
            self.client_measurement_counter[name] += 1
            
            self.overall_circuit.add_clbit(overall_qc_clbit)
            temp_qc.add_clbit(clbit)
            
            translation_dic[clbit] = overall_qc_clbit
        
        translation_dic.update({qb: qb for qb in qc.qubits})
        
        
        temp_qc.extend(qc)
        
        
        res, self.qs = single_shot_sim(temp_qc, self.qs)
        
        
        self.overall_circuit.barrier()
        self.overall_circuit.extend(qc, translation_dic)
            
        return {res: 1}
    
    def get_overall_qc(self):
        return self.overall_circuit.copy()
    
    def inbox(self, name):
        return self.inbox_dict[name]
    
    def reset_network_state(self):
        
        self.qs = QuantumState(0)
        self.qubit_map = {}
        self.overall_circuit = QuantumCircuit()
        self.client_measurement_counter = {}
        self.inbox = {}
        
def convert_qc_qubit_names(qc, name):
    
    res_qc = qc.clearcopy()
    
    res_qc.qubits = convert_qubit_names(qc.qubits, name)
    
    transl_dic = {qc.qubits[i].identifier : res_qc.qubits[i] for i in range(len(qc.qubits))}
    
    for instr in qc.data:
        res_qc.append(instr.op, [transl_dic[qb.identifier] for qb in instr.qubits], instr.clbits)
    
    return res_qc

def convert_qubit_names(qubits, name):
    
    res = []
    for i in range(len(qubits)):
        qb = qubits[i]
        if len(qb.identifier.split("@")) == 1:
            res.append(Qubit(qb.identifier + "@" + name ))
        else:
            res.append(Qubit(qb.identifier))
    
    return res
    
    

    
    
    


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-
import numpy as np   
from qrisp.circuit import transpile, QuantumCircuit, Instruction, Operation, ControlledOperation, fast_append
from qrisp.simulator.impure_quantum_state import ImpureQuantumState
from qrisp.simulator.quantum_state import QuantumState
from qrisp.simulator.circuit_preprocessing import circuit_preprocessor, count_measurements_and_treat_alloc, group_qc
from qrisp.interface.circuit_converter import convert_circuit
from tqdm import tqdm

import threading



#This dictionary contains the qrisp operations as values and their names as keys
#This is advantageous because during the simulation, we dont want to synthesize them every time they are required
# from qrisp.circuit import op_list
# op_dict = {op().name : op for op in op_list}


#This functions determines the quantum state after executing a quantum circuit
#and afterwards extracts the probability of measuring certain bit strings
def run(qc, shots, token = "", iqs = None, insert_reset = True):
    
    #This command enables fast appending. Fast appending means that the .append method of the QuantumCircuit
    #class checks much less validity conditions and is also less tolerant regarding inputs.
    
    print(f"Simulating {len(qc.qubits)} qubits..", end = "\r")
    
    
    with fast_append():
        
        #We convert the circuit which is given with portable objects to a qrisp circuit
        qc = convert_circuit(qc, "qrisp", transpile = True)
        
        
        #Count the amount of measurements (we can stop the simulation after all measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(qc, insert_reset = insert_reset)
        
        
        #Apply circuit preprocessing more
        qc = circuit_preprocessor(qc)
        
        
        measurement_counter = 0
        
        for i in range(len(qc.data)):
            if qc.data[i].op.name == "measure":
                measurement_counter += 1
            if measurement_counter == measurement_amount:
                break
            
        qc.data = qc.data[:i+1]
        
        
        if iqs == None:
            #Create impure quantum state object. This object tracks multiple decoherent quantum states that can appear when applying a non-unitary operation
            iqs = ImpureQuantumState(len(qc.qubits), clbit_amount = len(qc.clbits))
        
        #Counter to track how many measurements have been performed
        measurement_counter = 0
    
        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2**qc.data[i].op.num_qubits
        
        progress_bar = tqdm(
                            total = total_flops,
                            desc = f"Simulating {qc.num_qubits()} qubits..", 
                            bar_format="{desc} |{bar}| [{percentage:3.0f}%]", 
                            ncols = 85, 
                            # ascii = " >#",
                            leave = False,
                            delay = 0.2,
                            position = 0,
                            smoothing = 1
                            # colour = "green"
                            )
        
        
        #Main loop - this loop successively executes operations onto the impure quantum state object
        for i in range(len(qc.data)):
            
            #Set alias for the instruction of this operation        
            instr = qc.data[i]
    
            progress_bar.update(2**instr.op.num_qubits)
    
            
            #Gather the indices of the qubits from the circuits (ie. integers instead of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]
            
            
            #Perform instructions
            
            if instr.op.name == "reset":
                iqs.reset(qubit_indices[0])
            
            #Disentangling describes an operation, which mean that the superposition of two states can be safely
            #treated as two decoherent states. This is advantageous because if might be possible that the amplitude of one
            #state is 0, which means that we halfed the workload. Even if both states have non-zero amplitude, this still
            #yields an improvement because computing two decoherent states is more easily parallelized than the combined
            #coherent state
            elif instr.op.name == "disentangle":
                iqs.reset(qubit_indices[0], True)
            
            elif instr.op.name == "measure":
                iqs.measure(qubit_indices[0], qc.clbits.index(instr.clbits[0]))
                measurement_counter += 1
            
            elif instr.op.name[:4] == "c_if":
                iqs.conditionally_apply_operation(instr.op, qubit_indices, qc.clbits.index(instr.clbits[-1]))
            #If the operation is unitary, we apply this unitary on to the required qubit indices
            else:
                iqs.apply_operation(instr.op, qubit_indices)
            
    
            #If all measurements have been performed, break
            if measurement_counter == measurement_amount:
                break
        
        progress_bar.close()
        
        LINE_UP = '\033[1A'
        LINE_CLEAR = '\x1b[2K'
        print(LINE_CLEAR, end=LINE_UP)
        
        
        #Prepare result dictionary
        #The iqs object contains the outcome bitstrings in the attribute .outcome_list and the probablities in .cl_prob
        #In order to ensure qiskit compatibility, we reverse the bitstrings
        res = {}
        for i in range(len(iqs.cl_prob)):
            
            if iqs.cl_prob[i] == 0:
                continue
            
            outcome_str = ""
            for j in range(len(iqs.outcome_list[i])):
                if iqs.outcome_list[i][j]:
                    outcome_str += "1"
                else:
                    outcome_str += "0"
            
            try:
                res[outcome_str[::-1]] += int(np.round(iqs.cl_prob[i]*shots))
            except KeyError:
                res[outcome_str[::-1]] = int(np.round(iqs.cl_prob[i]*shots))
        
        return res


def statevector_sim(qc):
    
    #This command enables fast appending. Fast appending means that the .append method of the QuantumCircuit
    #class checks much less validity conditions and is also less tolerant regarding inputs.
    with fast_append():
        #We convert the circuit which is given with portable objects to a qrisp circuit
        
        if not isinstance(qc, QuantumCircuit):
            qc = convert_circuit(qc, "qrisp", transpile = True)
            
            
        qc = qc.copy()
        
        #TO-DO fix qubit ordering
        # qc.qubits = qc.qubits[::-1]
        
        #Count the amount of measurements (we can stop the simulation after all measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(qc, insert_reset = False)
        
        if measurement_amount != 0:
            raise Exception("Tried to determine the statevector of a circuit containing a measurement")
        
        #Apply circuit preprocessing more
        qc = group_qc(qc)
        
        if len(qc.data) == 0:
            res = np.zeros(2**len(qc.qubits), dtype = np.complex64)
            res[0] = 1
            return res
        
        qs = QuantumState(len(qc.qubits))
        
        
        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2**qc.data[i].op.num_qubits
        
        progress_bar = tqdm(
                            total = total_flops,
                            desc = f"Simulating {qc.num_qubits()} qubits..", 
                            bar_format="{desc} |{bar}| [{percentage:3.0f}%]", 
                            ncols = 85, 
                            #ascii = " >#",
                            leave = False,
                            delay = 0.2,
                            position = 0)
        
        
        
        
        #Main loop - this loop successively executes operations onto the impure quantum state object
        for i in range(len(qc.data)):
            
            #Set alias for the instruction of this operation        
            instr = qc.data[i]
                
            
            progress_bar.update(2**instr.op.num_qubits)
            #Gather the indices of the qubits from the circuits (ie. integers instead of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]
            
            #Perform instructions
            qs.apply_operation(instr.op, qubit_indices)
        
        res = qs.eval().tensor_array.to_array()
        
        progress_bar.close()
        
        LINE_UP = '\033[1A'
        LINE_CLEAR = '\x1b[2K'
        print(LINE_CLEAR, end=LINE_UP)
        
        
        
        #Deactivate the fast append mode
        QuantumCircuit.fast_append = False
        
        return res
    


def single_shot_sim(qc, quantum_state = None):
    
    
    if len(qc.data) == 0:
        return "", quantum_state
    
    #This command enables fast appending. Fast appending means that the .append method of the QuantumCircuit
    #class checks much less validity conditions and is also less tolerant regarding inputs.
    with fast_append():    
    
        
        #We convert the circuit which is given with portable objects to a qrisp circuit
        qc = convert_circuit(qc, "qrisp", transpile = True)
        
        
        
        
        #Treat allocation gates (ie. remove them)
        count_measurements_and_treat_alloc(qc, insert_reset = True)
        
        
        
        
        from qrisp.simulator import reorder_circuit    
        qc = group_qc(qc)
        
        
        qc = reorder_circuit(qc, ["measure", "reset", "disentangle"])
        
        
        if quantum_state == None:
            #Create quantum state object.
            quantum_state = QuantumState(len(qc.qubits))
        
        result_str = len(qc.clbits)*["0"]
        
        
        import random
        
        
        if len(qc.data):
            #Wrapper to pre calculate the unitaries from the preprocessed circuit in parallel with the main thread    
            def pre_calc_unitaries(op):
                try:
                    op.get_unitary()
                except:
                    pass
            
            # pre_calc_unitaries()
            
            pre_calc_thr = threading.Thread(target = pre_calc_unitaries, args = (qc.data[0].op,))
            pre_calc_thr.start()
            
        
        #Main loop - this loop successively executes operations onto the impure quantum state object
        for i in range(len(qc.data)):
            
            
            
            pre_calc_thr.join()
            
            if i < len(qc.data) - 1:
                
                pre_calc_thr = threading.Thread(target = pre_calc_unitaries, args = (qc.data[i+1].op,))
                pre_calc_thr.start()
            
            #Set alias for the instruction of this operation        
            instr = qc.data[i]
            
            #Gather the indices of the qubits from the circuits (ie. integers instead of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]
            
            
            
            # Perform instructions
            if instr.op.name == "reset":
                quantum_state.measure(qubit_indices[0])
                
                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome
                
                rng = random.random()
                
                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1
                
            
            #Disentangling describes an operation, which mean that the superposition of two states can be safely
            #treated as two decoherent states. This is advantageous because if might be possible that the amplitude of one
            #state is 0, which means that we halfed the workload. Even if both states have non-zero amplitude, this still
            #yields an improvement because computing two decoherent states is more easily parallelized than the combined
            #coherent state
            
            elif instr.op.name == "measure":
                
                quantum_state.measure(qubit_indices[0])
                
                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome
                
                
                rng = random.random()
                
                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1
                    result_str[qc.clbits.index(instr.clbits[0])] = "1"
                
            
            elif instr.op.name[:4] == "c_if":
                
                if result_str[qc.clbits.index(instr.clbits[0])] == "1":
                    quantum_state.apply_operation(instr.op, qubit_indices)
            #If the operation is unitary, we apply this unitary on to the required qubit indices
            else:
                quantum_state.apply_operation(instr.op, qubit_indices)
                
                
        return "".join(result_str)[::-1], quantum_state



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 06.05.2022
from qrisp.arithmetic import QuantumFloat, q_divmod
from qrisp.misc import multi_measurement

def test_quantum_divison():
    numerator = QuantumFloat(3, -2, signed = True)
    divisor = QuantumFloat(3, -2, signed = True)

    n = 4/8
    d = -5/4
    numerator.encode(n)
    divisor.encode(d)

    prec = 4
    quotient, remainder = q_divmod(numerator, divisor, prec = prec, adder = "thapliyal")

    q, r = list(multi_measurement([quotient, remainder]))[0]

    print("Q: ", q)
    print("delta_Q: ", abs(q-n/d))
    print("R: ", r)
    assert q == -0.375 and r == 0.03125 and abs(q-n/d) < 2**-prec



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

from qrisp.simulator.tensor_factor import TensorFactor, multi_entangle
from qrisp.simulator.numerics_config import xp

#This class describes a quantum state
#This is achieved by another class, the TensorFactor.
#Each qubit initially has its own tensor factor. When an entangling operation is executed,
#the TensorFactors can be entangled. We then have the situation that two qubits are described by the same TensorFactor

#This structure allows the simulation to perform entangling operations only when they are really needed
#If a circuit consists of two blocks which are never entangled, after the simulation, the QuantumState
#will consist of two TensorFactors
class QuantumState:
    
    def __init__(self, n):
        
        #The index of the tensor factors refers to what qubit they are describing
        self.tensor_factors = [TensorFactor([i]) for i in range(n)]
        self.n = n
    
    #This functions applies a (unitary) operation onto the tensor factor
    def apply_operation(self, operation, qubits):
        #Determine the involved tensor factors and the involved qubits        
        involved_factors = list(set(sum([[self.tensor_factors[qubits[i]]] for i in range(len(qubits))], [])))
        involved_qubits = list(set(sum([tf.qubits for tf in involved_factors], [])))
        
        #Entangle all the factors involved in the operation
        entangled_factor = multi_entangle(involved_factors)
        
        #Update the entangled factor on the involved qubits
        for i in range(len(involved_qubits)):
            self.tensor_factors[involved_qubits[i]] = entangled_factor
        
        #Get the unitary of the operation
        unitary = operation.get_unitary()
        # unitary = operation.get
        
        #Apply the matrix
        entangled_factor.apply_matrix(unitary, qubits)
    
    #Method to retrieve the statevector in the form of an array of complex numbers
    def eval(self):
        
        [factor.unravel() for factor in self.tensor_factors]
        
        #print(len(list(set(self.tensor_factors))))
        
        entangled_factor = multi_entangle(list(set(self.tensor_factors)))
        
        entangled_factor.unravel()
        return entangled_factor
    
    #Copying method
    def copy(self):
        
        factor_list = list(set(self.tensor_factors))
        new_factor_list = [factor.copy() for factor in factor_list]
        
        res = QuantumState(self.n)
        
        for i in range(len(new_factor_list)):
            factor_qubits = new_factor_list[i].qubits
            for j in range(len(factor_qubits)):
                res.tensor_factors[factor_qubits[j]] = new_factor_list[i]
        
        return res
    
    
    #Method to perform a measurement on this quantum state
    def measure(self, i, keep_res = True):
        
        #Determine the TensorFactor to be measured
        tensor_factor = self.tensor_factors[i]
        
        #Perform measurement on this TensorFactor
        #p_0 is the measurement probability of measuring p_0,
        #outcome_tensor_f_0 is the TensorFactor that is left after a measurement with outcome 0 has been performed
        #ie. if the state descibes |psi> = 2**-0.5 |0>|a> + 2**-0.5|1>|b>
        #we have p_0 = 0.5 and outcome_tensor_f_0 = |a>
        p_0, outcome_tensor_f_0, p_1, outcome_tensor_f_1 = tensor_factor.measure(i)
        
        
        #This code block handles several cases that mainly serve to prevent simulation of QuantumStates with
        #a probablity of 0 to be measured.
        if p_0 != 0:
            
            #If both measurement results have non-zero probability, we copy self
            #In case only one state has non-zero probability we can continue using self
            if p_1 != 0:
                outcome_state_0 = self.copy()
            else:
                outcome_state_0 = self
            
            #Update the tensor factors of the qubits that have been entangled to the measured qubit
            outcome_qubits = outcome_tensor_f_0.qubits
            for j in range(len(outcome_qubits)):
                outcome_state_0.tensor_factors[outcome_qubits[j]] = outcome_tensor_f_0
            
            #Update the qubit that has been measured (now it's own tensor factor again)
            outcome_state_0.tensor_factors[i] = TensorFactor([i])
            
        else:
            outcome_state_0 = None
        
        #In case the probability to measure a 1 is non-zero, we can basically proceed as above
        #but we need to set the measured qubit to the |1> state if this is required
        if p_1 != 0:
            outcome_state_1 = self
            temp = outcome_tensor_f_1.qubits
            for j in range(len(temp)):
                outcome_state_1.tensor_factors[temp[j]] = outcome_tensor_f_1
            
            if keep_res:
                #This set the measured qubit to the |1> state (described by the array [0,1] instead of [1,0] which descibes the |0> state)
                outcome_state_1.tensor_factors[i] = TensorFactor([i], xp.array([0,1], dtype = xp.complex64))
            else:
                outcome_state_1.tensor_factors[i] = TensorFactor([i])
        else:
            outcome_state_1 = None
        
        #log measurement out come. Note that we cant return here, because this functions is called as a thread, implying it does not keep the returned result
        self.last_mes_outcome = (p_0, outcome_state_0, p_1, outcome_state_1)
        
        return self.last_mes_outcome
    
    def add_qubit(self):
        
        self.tensor_factors.append(TensorFactor([self.n]))
        self.n += 1


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 05.05.2022
import pytest
import numpy as np

from qrisp import QuantumSession, QuantumVariable
from qrisp.misc.GMS_tools import gms_multi_cx_fan_out, gms_multi_cx_fan_in, gms_multi_cp_gate, gms_multi_cp_gate_mono_phase, GXX_wrapper
from qrisp import x

def test_GXX_gates_example():
    n = 5
    qs = QuantumSession()
    qv = QuantumVariable(n, qs)

    x(qv[-1])
    qs.append(gms_multi_cx_fan_out(n - 1, use_uniform=True, phase_tolerant=False, basis = 'GXX'), qv.reg)

    
    tmp = qv.get_measurement()
    
    print(tmp)
    assert qv.get_measurement() == {'11111': 1.0}



    ###################
    n = 3

    theta = [np.pi / (i + 1) for i in range(n - 1)]
    qs_0 = QuantumSession()
    qv = QuantumVariable(n, qs_0)
    qs_0.append(gms_multi_cp_gate(n - 1, theta), qv.reg)
    
    qs_1 = QuantumSession()
    qv = QuantumVariable(n, qs_1)

    for i in range(n - 1):
        qs_1.cp(theta[i], i, n - 1)

    unitary = qs_0.get_unitary()
    print(qs_0.get_unitary(decimals=4))
    tmparray = qs_0.get_unitary(decimals=4)
    assert type(qs_0.get_unitary(decimals=4)).__module__ == 'numpy'
    while i < len(tmparray):
        assert all(tmparray[i]) in [1. + 0.j, -0. - 0.j, -0. + 0.j, 0. + 0.j, 1. - 0.j, 0. - 0.j, 0. + 1.j, -1. + 0.j,
                                    -0. - 1.j]
        assert not bool(all(tmparray[i]) == 7153)
        i += 1

    print("Is equal:", qs_0.compare_unitary(qs_1))
    assert qs_0.compare_unitary(qs_1)

    print("Is diagonal:", np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1E-3)
    assert np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1E-3

    print(np.round(np.angle(np.diagonal(unitary)) % (2 * np.pi) / (np.pi), 3))
